from collections.abc import Iterable
from typing import NamedTuple

import quantiphy
from quantiphy import Quantity

# to make unit conversion work, unit must be a non-empty string
_SENTINEL_UNIT = "SENTINEL_UNIT"


class PrettyQuantitiesComponents(NamedTuple):
    mantissas: list[str]
    spacer: str
    units: str


class PrettyQuantityComponents(NamedTuple):
    whole: str
    frac: str
    units: str


def pretty_quantities(
    values: Iterable[float], unit: str | None = None, *, prec: int = 2, **kwargs
) -> PrettyQuantitiesComponents:
    """.

    Examples:
        >>> pretty_quantities([0.1234, 0.01234, 0.001234], "s")
        PrettyQuantitiesComponents(mantissas=['123.', '12.', '1.'], spacer=' ', units='ms')
    """
    unit_or_sentinel: str = unit or _SENTINEL_UNIT
    components: PrettyQuantityComponents = pretty_quantity_components(
        max(values), unit_or_sentinel, prec=prec, **kwargs
    )
    spacer: str = _get_spacer(unit)

    def number_fmt(whole: str, frac: str, _units: str) -> str:
        if not frac.startswith("."):
            frac = f".{frac}"
        return f"{whole}{frac}"

    with Quantity.prefs(
        map_sf=Quantity.map_sf_to_greek,
        number_fmt=number_fmt,
        strip_radix=False,
        strip_zeros=False,
        show_units=False,
    ):
        mantissas: list[str] = [
            quantiphy.fixed(
                value,
                unit_or_sentinel,
                prec=len(components.frac) - 1,
                scale=components.units,
            )
            for value in values
        ]
    units: str = components.units
    if unit is None:
        units = units.removesuffix(_SENTINEL_UNIT)
    return PrettyQuantitiesComponents(mantissas, spacer, units)


def pretty_quantity(
    value: float, unit: str | None = None, *, prec: int = 2, **kwargs
) -> str:
    """.

    Examples:
        >>> pretty_quantity(0.1234, "s")
        '123. ms'
        >>> pretty_quantity(0.01234, "s")
        '12.3 ms'
    """
    spacer: str = _get_spacer(unit)

    def number_fmt(whole: str, frac: str, units: str) -> str:
        if not frac.startswith("."):
            frac = f".{frac}"
        mantissa: str = f"{whole}{frac}"
        return f"{mantissa}{spacer}{units}"

    with Quantity.prefs(
        map_sf=Quantity.map_sf_to_greek,
        number_fmt=number_fmt,
        strip_radix=False,
        strip_zeros=False,
    ):
        q: Quantity = Quantity(value) if unit is None else Quantity(value, unit)
        return q.render(prec=prec, **kwargs)


def pretty_quantity_components(
    value: float, unit: str | None = None, *, prec: int = 2, **kwargs
) -> PrettyQuantityComponents:
    number_fmt = _NumberFmt()
    with Quantity.prefs(
        map_sf=Quantity.map_sf_to_greek,
        number_fmt=number_fmt,
        strip_radix=False,
        strip_zeros=False,
    ):
        quantiphy.render(value, unit or _SENTINEL_UNIT, prec=prec, **kwargs)
    whole: str = number_fmt.whole
    frac: str = number_fmt.frac
    units: str = number_fmt.units
    if not frac.startswith("."):
        frac = f".{frac}"
    if unit is None:
        units = units.removesuffix(_SENTINEL_UNIT)
    return PrettyQuantityComponents(whole, frac, units)


class _NumberFmt:
    whole: str
    frac: str
    units: str

    def __call__(self, whole: str, frac: str, units: str) -> str:
        self.whole = whole
        self.frac = frac
        self.units = units
        return ""


def _get_spacer(unit: str | None) -> str:
    if not unit:
        return ""
    if unit in Quantity.get_pref("tight_units"):
        return ""
    return Quantity.get_pref("spacer")



from __future__ import annotations

import linecache
import types
from typing import TYPE_CHECKING, Any, Self, override

import executing

if TYPE_CHECKING:
    from _typeshed import StrPath


class Source(executing.Source):
    if TYPE_CHECKING:

        @override
        @classmethod
        def for_frame(cls, frame: types.FrameType, use_cache: bool = True) -> Self: ...

    @override
    @classmethod
    def for_filename(
        cls,
        filename: StrPath,
        module_globals: dict[str, Any] | None = None,
        use_cache: bool = True,
    ) -> Self:
        # Intentionally skip linecache.checkcache() to preserve consistent source code.
        # This ensures icecream displays correct source lines and variable names, since
        # we don't support hot reloading and need stable source references.
        filename = str(filename)
        return cls._for_filename_and_lines(
            filename, tuple(linecache.getlines(filename, module_globals))
        )  # pyright: ignore[reportReturnType]



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, overload

from ._decode import DecHook, PydanticValidateOptions
from ._serde import json, toml, yaml

if TYPE_CHECKING:
    from _typeshed import StrPath

readers: dict[str, Callable] = {}
readers[".json"] = json.load
readers[".toml"] = toml.load
readers[".yaml"] = yaml.load
readers[".yml"] = yaml.load


@overload
def load(
    path: StrPath,
    /,
    *,
    dec_hook: DecHook | None = ...,
    force_ext: str | None = None,
    pydantic: PydanticValidateOptions | None = None,
    strict: bool = True,
) -> Any: ...
@overload
def load[T](
    path: StrPath,
    /,
    *,
    dec_hook: DecHook | None = ...,
    force_ext: str | None = None,
    pydantic: PydanticValidateOptions | None = None,
    strict: bool = True,
    type: type[T],
) -> T: ...
@overload
def load[T](
    path: StrPath,
    /,
    *,
    dec_hook: DecHook | None = ...,
    force_ext: str | None = None,
    pydantic: PydanticValidateOptions | None = None,
    strict: bool = True,
    type: Any,
) -> Any: ...
def load(path: StrPath, /, *, force_ext: str | None = None, **kwargs) -> Any:
    path = Path(path)
    ext: str = force_ext or path.suffix
    return readers[ext](path, **kwargs)



import attrs
from rich.syntax import Syntax, SyntaxTheme

from liblaf.grapes._config import config


@attrs.define
class RichTracebackOptions:
    indent_guide: bool = attrs.field(factory=config.traceback.indent_guide.get)
    locals_hide_dunder: bool = attrs.field(
        factory=config.traceback.locals_hide_dunder.get
    )
    locals_hide_sunder: bool = attrs.field(
        factory=config.traceback.locals_hide_sunder.get
    )
    locals_max_length: int = attrs.field(factory=config.traceback.locals_max_length.get)
    locals_max_string: int = attrs.field(factory=config.traceback.locals_max_string.get)
    show_locals: bool = attrs.field(factory=config.traceback.show_locals.get)
    theme: SyntaxTheme = attrs.field(
        factory=lambda: Syntax.get_theme(config.traceback.theme.get())
    )



import contextlib
from collections.abc import Generator

import attrs


@attrs.define
class MissingOptionalDependencyError(ImportError):
    err: ImportError
    package: str | None = None
    extra: str | None = None

    def __str__(self) -> str:
        msg: str = str(self.err)
        if self.package and self.extra:
            msg += f"\nPlease install optional dependencies with '{self.package}[{self.extra}]'"
        return msg


@contextlib.contextmanager
def optional_deps(
    package: str | None = None, extra: str | None = None
) -> Generator[None]:
    try:
        yield
    except ImportError as err:
        raise MissingOptionalDependencyError(
            err, package=package, extra=extra
        ) from None



from ._abc import RichHandlerColumn
from ._level import RichHandlerColumnLevel
from ._location import RichHandlerColumnLocation
from ._time import RichHandlerColumnTime

__all__ = [
    "RichHandlerColumn",
    "RichHandlerColumnLevel",
    "RichHandlerColumnLocation",
    "RichHandlerColumnTime",
]



from typing import Any, Unpack

import wadler_lindig as wl
from rich.repr import RichReprResult

from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions

from ._type import pdoc_type


def pdoc_rich_repr(
    obj: Any, **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc | None:
    if not hasattr(obj, "__rich_repr__"):
        return None
    repr_result: RichReprResult | None = obj.__rich_repr__()
    if repr_result is None:
        return None
    cls: type = type(obj)
    args: list[Any] = []
    pairs: list[tuple[str, Any]] = []
    for field in repr_result:
        name: str
        value: Any
        default: Any
        if isinstance(field, tuple):
            if len(field) == 2:
                name, value = field
                pairs.append((name, value))
            elif len(field) == 3:
                name, value, default = field
                if kwargs.get("hide_defaults", True) and value is default:
                    continue
                pairs.append((name, value))
        else:
            value = field
            args.append(value)
    return wl.bracketed(
        begin=pdoc_type(cls, dataclass=True, **kwargs) + wl.TextDoc("("),
        docs=[wl.pdoc(arg, **kwargs) for arg in args] + wl.named_objs(pairs, **kwargs),
        sep=wl.comma,
        end=wl.TextDoc(")"),
        indent=kwargs.get("indent", 2),
    )



from rich.console import RenderableType
from rich.progress import ProgressColumn, Task
from rich.table import Column
from rich.text import Text

from liblaf.grapes import pretty


class RateColumn(ProgressColumn):
    unit: str = "it"

    def __init__(self, unit: str = "it", table_column: Column | None = None) -> None:
        super().__init__(table_column)
        self.unit = unit

    def render(self, task: Task) -> RenderableType:
        if not task.speed:
            return Text(f"?{self.unit}/s", style="progress.data.speed")
        throughput: str = pretty.pretty_throughput(task.speed, self.unit)
        return Text(throughput, style="progress.data.speed")



from collections.abc import Iterable

from ._quantiphy import PrettyQuantitiesComponents, pretty_quantities, pretty_quantity


def pretty_duration(seconds: float, *, prec: int = 2, **kwargs) -> str:
    """.

    Examples:
        >>> pretty_duration(0.1234)
        '123. ms'
        >>> pretty_duration(0.01234)
        '12.3 ms'
    """
    return pretty_quantity(seconds, "s", prec=prec, **kwargs)


def pretty_durations(
    seconds: Iterable[float], *, prec: int = 2, **kwargs
) -> PrettyQuantitiesComponents:
    """.

    Examples:
        >>> pretty_durations([0.1234, 0.01234, 0.001234])
        PrettyQuantitiesComponents(mantissas=['123.', '12.', '1.'], spacer=' ', units='ms')
    """
    return pretty_quantities(seconds, "s", prec=prec, **kwargs)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from liblaf.grapes._config import config
from liblaf.grapes.rich.logging import RichHandler

from .filters import LimitsFilter
from .handlers import RichFileHandler
from .helpers import (
    init_levels,
    install_excepthook,
    install_unraisablehook,
    remove_non_root_stream_handlers,
    set_default_logger_level_by_release_type,
)

if TYPE_CHECKING:
    from _typeshed import StrPath


def init(
    *,
    file: StrPath | None = None,
    force: bool = False,
    time_relative: bool | None = None,
) -> None:
    """.

    Examples:
        >>> init()
    """
    if file is None:
        file = config.logging.file.get()
    handlers: list[logging.Handler] = []
    if force or not logging.root.hasHandlers():
        handlers.append(RichHandler(time_relative=time_relative))
        if file is not None:
            handlers.append(RichFileHandler(file))
        for handler in handlers:
            handler.addFilter(LimitsFilter())
    init_levels()
    install_excepthook()
    install_unraisablehook()
    logging.basicConfig(handlers=handlers, force=force)
    logging.captureWarnings(True)  # noqa: FBT003
    logging.getLogger("IPKernelApp").setLevel(logging.WARNING)
    logging.getLogger("liblaf").setLevel(logging.DEBUG)
    remove_non_root_stream_handlers()
    set_default_logger_level_by_release_type()



from __future__ import annotations

import logging
import sys

logger: logging.Logger = logging.getLogger()


def install_unraisablehook(level: int = logging.ERROR) -> None:
    def unraisablehook(args: sys.UnraisableHookArgs, /) -> None:
        if args.exc_value is None:
            return
        logger.log(
            level,
            "%s: %r",
            args.err_msg,
            args.object,
            exc_info=(args.exc_type, args.exc_value, args.exc_traceback),
        )

    sys.unraisablehook = unraisablehook



from typing import Unpack

import wadler_lindig as wl
from wadler_lindig._definitions import _pformat_type

from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions


def pdoc_type(
    obj: type, *, dataclass: bool = False, **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc:
    if dataclass:
        kwargs["show_type_module"] = kwargs.get("show_dataclass_module", False)
    return _pformat_type(obj, **kwargs)



import about_time


def pretty_throughput(value: float, unit: str = "") -> str:
    """.

    Examples:
        >>> pretty_throughput(1234, "it")
        '1.2kit/s'
    """
    throughput: about_time.HumanThroughput = about_time.HumanThroughput(value, unit)
    return throughput.as_human()



import logging
import sys
import types

logger: logging.Logger = logging.getLogger()


def install_excepthook(level: int = logging.CRITICAL) -> None:
    def excepthook(
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: types.TracebackType | None,
    ) -> None:
        logger.log(level, exc_value, exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = excepthook



from ._abbr_path import abbr_path
from ._ci import in_ci
from ._entrypoint import entrypoint
from ._frame import (
    get_frame,
    get_frame_with_stacklevel,
    hidden_from_logging,
    hidden_from_traceback,
    hidden_from_warnings,
)
from ._release_type import is_dev_release, is_pre_release

__all__ = [
    "abbr_path",
    "entrypoint",
    "get_frame",
    "get_frame_with_stacklevel",
    "hidden_from_logging",
    "hidden_from_traceback",
    "hidden_from_warnings",
    "in_ci",
    "is_dev_release",
    "is_pre_release",
]



import sys
import types

from rich.console import Console

from liblaf.grapes.rich._get_console import get_console

from ._exception import RichExceptionSummary


def install(*, console: Console | None = None) -> None:
    if console is None:
        console = get_console(stderr=True)

    def excepthook(
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: types.TracebackType | None,
    ) -> None:
        exc = RichExceptionSummary(exc_type, exc_value, traceback)
        console.print(exc)

    sys.excepthook = excepthook



import enum
from typing import Unpack

import wadler_lindig as wl

from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions

from ._type import pdoc_type


def pdoc_enum(obj: enum.Enum, **kwargs: Unpack[WadlerLindigOptions]) -> wl.AbstractDoc:
    return pdoc_type(type(obj), dataclass=True, **kwargs) + wl.TextDoc(f".{obj.name}")



import functools
import types
from collections.abc import Generator
from typing import Self

import attrs
from rich.console import Console, ConsoleOptions, Group, RenderableType, RenderResult
from rich.highlighter import Highlighter, ReprHighlighter
from rich.panel import Panel
from rich.text import Text

from liblaf.grapes import pretty

from ._options import RichTracebackOptions
from ._stack import RichStackSummary


@attrs.define
class RichExceptionSummary:
    exc_type: type[BaseException]
    exc_value: BaseException
    traceback: types.TracebackType | None
    highlighter: Highlighter = attrs.field(
        repr=False, init=False, factory=ReprHighlighter
    )
    _seen: set[int] = attrs.field(factory=set, repr=False, alias="_seen")

    def __attrs_post_init__(self) -> None:
        self._seen.add(id(self.exc_value))

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield from self.render()

    @functools.cached_property
    def cause(self) -> Self | None:
        cause: BaseException | None = self.exc_value.__cause__
        if cause is not None and id(cause) not in self._seen:
            return type(self)(type(cause), cause, cause.__traceback__, _seen=self._seen)
        return None

    @functools.cached_property
    def context(self) -> Self | None:
        context: BaseException | None = self.exc_value.__context__
        if (
            context is not None
            and not self.exc_value.__suppress_context__
            and id(context) not in self._seen
        ):
            return type(self)(
                type(context), context, context.__traceback__, _seen=self._seen
            )
        return None

    @functools.cached_property
    def exceptions(self) -> list[Self]:
        if not isinstance(self.exc_value, (BaseExceptionGroup, ExceptionGroup)):
            return []
        exceptions: list[Self] = []
        for exception in self.exc_value.exceptions:
            if id(exception) in self._seen:
                continue
            exceptions.append(
                type(self)(
                    type(exception),
                    exception,
                    exception.__traceback__,
                    _seen=self._seen,
                )
            )
        return exceptions

    @functools.cached_property
    def exc_type_str(self) -> str:
        module: str = self.exc_type.__module__
        qualname: str = self.exc_type.__qualname__
        if module in ("__main__", "builtins"):
            return qualname
        return f"{module}.{qualname}"

    @functools.cached_property
    def stack(self) -> RichStackSummary:
        return RichStackSummary(self.traceback)

    def render(
        self, options: RichTracebackOptions | None = None
    ) -> Generator[RenderableType]:
        if options is None:
            options = RichTracebackOptions()
        if self.cause is not None:
            yield from self.cause.render(options)
            yield Text(
                "\nThe above exception was the direct cause of the following exception:\n",
                style="italic",
            )
        if self.context is not None:
            yield from self.context.render(options)
            yield Text(
                "\nDuring handling of the above exception, another exception occurred:\n",
                style="italic",
            )
        yield from self.stack.render(options)
        yield from self._render_exception_only(options)
        if self.exceptions:
            yield from self._render_subexceptions(options)

    def _render_exception_only(
        self, _options: RichTracebackOptions
    ) -> Generator[RenderableType]:
        exc_value: str = str(self.exc_value)
        if exc_value:
            exc_value_text: Text
            if pretty.has_ansi(exc_value):
                exc_value_text = Text.from_ansi(exc_value, style="traceback.exc_value")
            else:
                exc_value_text = Text(exc_value, style="traceback.exc_value")
                exc_value_text = self.highlighter(exc_value_text)
            yield Text.assemble(
                (f"{self.exc_type_str}:", "traceback.exc_type"), " ", exc_value_text
            )
        else:
            yield Text(self.exc_type_str, style="traceback.exc_type")
        for note in getattr(self.exc_value, "__notes__", ()):
            yield Text.assemble(("[NOTE]", "traceback.note"), " ", note)

    def _render_subexceptions(
        self, options: RichTracebackOptions
    ) -> Generator[RenderableType]:
        for i, exception in enumerate(self.exceptions, start=1):
            panel = Panel(
                Group(*exception.render(options)),
                title=f"Sub-exception #{i}",
                expand=False,
                border_style="traceback.group.border",
            )
            yield panel



from typing import Any, Unpack

import wadler_lindig as wl

from liblaf.grapes.functools import wraps

from ._options import WadlerLindigOptions, make_kwargs


@wraps(wl.pdoc)
def pdoc(obj: Any, **kwargs: Unpack[WadlerLindigOptions]) -> wl.AbstractDoc:
    kwargs: WadlerLindigOptions = make_kwargs(**kwargs)
    return wl.pdoc(obj, **kwargs)



from ._array import pdoc_array
from ._chain import chain_custom
from ._datetime import pdoc_datetime
from ._dispatch import PdocCustomDispatcher, pdoc_custom
from ._enum import pdoc_enum
from ._fieldz import pdoc_fieldz
from ._mapping import pdoc_mapping
from ._pydantic import pdoc_pydantic_root_model
from ._rich_repr import pdoc_rich_repr
from ._type import pdoc_type

__all__ = [
    "PdocCustomDispatcher",
    "chain_custom",
    "pdoc_array",
    "pdoc_custom",
    "pdoc_datetime",
    "pdoc_enum",
    "pdoc_fieldz",
    "pdoc_mapping",
    "pdoc_pydantic_root_model",
    "pdoc_rich_repr",
    "pdoc_type",
]



import enum
from typing import Final, Literal


class _Missing(enum.Enum):
    MISSING = enum.auto()

    def __repr__(self) -> str:
        return f"<{self.name}>"


type MissingType = Literal[_Missing.MISSING]
MISSING: Final[MissingType] = _Missing.MISSING



import functools
import types
from collections.abc import Generator

import attrs
import rich
from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.panel import Panel
from rich.text import Text

from ._frame import RichFrameSummary
from ._options import RichTracebackOptions


@attrs.define
class RichStackSummary:
    traceback: types.TracebackType | None

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield from self.render()

    @functools.cached_property
    def frames(self) -> list[RichFrameSummary]:
        frames: list[RichFrameSummary] = []
        traceback: types.TracebackType | None = self.traceback
        while traceback is not None:
            frames.append(
                RichFrameSummary(
                    traceback.tb_frame, traceback.tb_lineno, traceback.tb_lasti
                )
            )
            traceback = traceback.tb_next
        return frames

    def render(
        self, options: RichTracebackOptions | None = None
    ) -> Generator[RenderableType]:
        if options is None:
            options = RichTracebackOptions()
        if not self.frames:
            return
        panel = Panel(
            self._render_frames(options),
            title=Text.assemble(
                "Traceback ",
                ("(most recent call last)", "dim"),
                style="traceback.title",
            ),
            expand=False,
            style=options.theme.get_background_style(),
            border_style="traceback.border",
        )
        yield panel

    @rich.console.group()
    def _render_frames(self, options: RichTracebackOptions) -> RenderResult:
        for frame in self.frames:
            yield from frame.render(options)



import functools
from collections.abc import Callable, Iterable
from typing import Any, overload


@overload
def wraps[C: Callable](  # pyright: ignore[reportInconsistentOverload]
    wrapped: C,
    assigned: Iterable[str] = functools.WRAPPER_ASSIGNMENTS,
    updated: Iterable[str] = functools.WRAPPER_UPDATES,
) -> Callable[[Any], C]: ...
def wraps[C: Callable](wrapped: C, *args, **kwargs) -> Callable[[Any], C]:
    return functools.wraps(wrapped, *args, **kwargs)  # pyright: ignore[reportReturnType]



import functools
import logging
import types
from typing import Any

from liblaf.grapes import magic

_PROXY_METHODS: set[str] = {
    "debug",
    "info",
    "warning",
    "error",
    "critical",
    "log",
    "exception",
}


class AutoLogger:
    def __getattr__(self, name: str) -> Any:
        if name in _PROXY_METHODS:
            return functools.partial(self._delegate, name)
        frame: types.FrameType | None = magic.get_frame(depth=2)
        logger_name: str | None = None
        if frame is not None:
            logger_name = frame.f_globals.get("__name__")
        logger: logging.Logger = logging.getLogger(logger_name)
        return getattr(logger, name)

    def _delegate(self, method: str, *args, **kwargs) -> Any:
        depth: int = kwargs.get("stacklevel", 1)
        frame: types.FrameType | None
        stacklevel: int
        frame, stacklevel = magic.get_frame_with_stacklevel(
            depth=depth + 1, hidden=magic.hidden_from_logging
        )
        logger_name: str | None = None
        if frame is not None:
            logger_name = frame.f_globals.get("__name__")
        logger: logging.Logger = logging.getLogger(logger_name)
        kwargs["stacklevel"] = stacklevel
        return getattr(logger, method)(*args, **kwargs)


autolog = AutoLogger()



from . import columns
from ._handler import RichHandler
from .columns import (
    RichHandlerColumn,
    RichHandlerColumnLevel,
    RichHandlerColumnLocation,
    RichHandlerColumnTime,
)

__all__ = [
    "RichHandler",
    "RichHandlerColumn",
    "RichHandlerColumnLevel",
    "RichHandlerColumnLocation",
    "RichHandlerColumnTime",
    "columns",
]



from typing import Unpack

import pydantic
import wadler_lindig as wl

from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions

from ._type import pdoc_type


def pdoc_pydantic_root_model(
    obj: pydantic.RootModel, **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc:
    return pdoc_type(type(obj), dataclass=True, **kwargs) + wl.pdoc(obj.root, **kwargs)



from __future__ import annotations

import functools
from collections.abc import Callable, Iterable
from typing import Any, TypedDict, Unpack, overload

import attrs
import environs

from liblaf.grapes.sentinel import MISSING, MissingType

from ._constants import METADATA_KEY
from ._entry import Entry
from ._field import Field


# I don't know why but mkdocstrings will fail if I use
# `environs.types.BaseMethodKwargs` directly, so I redefine it here as a
# workaround.
class BaseMethodKwargs(TypedDict, total=False):
    # kwargs shared by all parser methods
    validate: Callable[[Any], Any] | Iterable[Callable[[Any], Any]] | None


@attrs.define
class VarEntry[T](Entry[Field[T]]):
    getter: Callable[[str], T]
    default: T | MissingType = MISSING
    env: str | None = None
    factory: Callable[[], T] | None = None

    def make(self, field: attrs.Attribute, prefix: str) -> Field[T]:
        return Field(
            name=prefix + field.name,
            default=self.default,
            env=self.env,
            factory=self.factory,
            getter=self.getter,
        )


@attrs.define
class FieldMethod[T]:
    wrapped: environs.FieldMethod[T]

    @overload
    def __call__(
        self, *, default: T, env: str | None = None, **kwargs: Unpack[BaseMethodKwargs]
    ) -> Field[T]: ...
    @overload
    def __call__(
        self,
        *,
        default: None,
        env: str | None = None,
        **kwargs: Unpack[BaseMethodKwargs],
    ) -> Field[T | None]: ...
    @overload
    def __call__(
        self,
        *,
        factory: Callable[[], T],
        env: str | None = None,
        **kwargs: Unpack[BaseMethodKwargs],
    ) -> Field[T]: ...
    @overload
    def __call__(
        self,
        *,
        env: str | None = None,
        factory: None = None,
        **kwargs: Unpack[BaseMethodKwargs],
    ) -> Field[T]: ...
    def __call__(
        self,
        *,
        default: T | MissingType | None = MISSING,
        env: str | None = None,
        factory: Callable[[], T] | None = None,
        **kwargs: Unpack[BaseMethodKwargs],
    ) -> Field[T] | Field[T | None]:
        return attrs.field(
            metadata={
                METADATA_KEY: VarEntry(
                    getter=functools.partial(self.wrapped, **kwargs),
                    default=default,
                    factory=factory,
                    env=env,
                )
            }
        )


@attrs.define
class ListFieldMethod[T]:
    wrapped: environs.ListFieldMethod

    def __call__(
        self,
        subcast: environs.Subcast[T] | None = None,
        *,
        delimiter: str | None = None,
        env: str | None = None,
        factory: Callable[[], list[T]] | None = list,
        **kwargs: Unpack[BaseMethodKwargs],
    ) -> Field[list[T]]:
        return attrs.field(
            metadata={
                METADATA_KEY: VarEntry(
                    getter=functools.partial(
                        self.wrapped, subcast=subcast, delimiter=delimiter, **kwargs
                    ),
                    factory=factory,
                    env=env,
                )
            }
        )



from typing import Any, Unpack, cast

import tlz
import wadler_lindig as wl

from liblaf.grapes._config import config

from ._typing import WadlerLindigOptions
from .custom import chain_custom, pdoc_custom


def make_kwargs(**kwargs: Unpack[WadlerLindigOptions]) -> WadlerLindigOptions:
    kwargs: dict[str, Any] = tlz.merge(config.pretty.get(), kwargs)
    prelude_custom = _PreludeCustom(kwargs)
    kwargs["custom"] = chain_custom(kwargs.get("custom"), prelude_custom)
    return cast("WadlerLindigOptions", kwargs)


class _PreludeCustom:
    kwargs: dict[str, Any]  # keep a reference to use updated kwargs

    def __init__(self, kwargs: dict[str, Any]) -> None:
        self.kwargs = kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> wl.AbstractDoc | None:
        kwargs: dict[str, Any] = {**self.kwargs, **kwargs}
        return pdoc_custom(*args, **kwargs)



from logging import LogRecord
from typing import override

import attrs
from pendulum import DateTime
from rich.text import Text

from liblaf.grapes._config import config

from ._abc import RichHandlerColumn


@attrs.define
class RichHandlerColumnTime(RichHandlerColumn):
    fmt: str = attrs.field(factory=config.logging.datefmt.get)
    relative: bool = attrs.field(factory=config.logging.time_relative.get)

    @override
    def render(self, record: LogRecord) -> Text:
        plain: str = (
            self._render_relative(record)
            if self.relative
            else self._render_absolute(record)
        )
        return Text(plain, "log.time")

    def _render_absolute(self, record: LogRecord) -> str:
        time: DateTime = DateTime.fromtimestamp(record.created)
        return time.format(self.fmt)

    def _render_relative(self, record: LogRecord) -> str:
        milliseconds: int = round(record.relativeCreated)
        days: int
        days, milliseconds = divmod(milliseconds, 86400000)
        hours: int
        hours, milliseconds = divmod(milliseconds, 3600000)
        minutes: int
        minutes, milliseconds = divmod(milliseconds, 60000)
        seconds: int
        seconds, milliseconds = divmod(milliseconds, 1000)
        text: str = f"{hours:02}:{minutes:02}:{seconds:02}.{milliseconds:03}"
        return f"{days}d,{text}" if days > 0 else text



from __future__ import annotations

from collections.abc import Container, Iterable
from typing import TYPE_CHECKING

from liblaf.grapes import warnings

if TYPE_CHECKING:
    from _typeshed import SupportsGetItem


_DEPRECATED_MESSAGE = "'{deprecated_key}' is deprecated. Please use '{key}' instead."


def contains[T](
    obj: Container[T],
    key: T,
    deprecated_keys: Iterable[T] = (),
    *,
    msg: str = _DEPRECATED_MESSAGE,
) -> bool:
    """.

    Examples:
        >>> import pytest
        >>> data = {"a": 1, "b": 2}
        >>> contains(data, "a")
        True
        >>> contains(data, "missing")
        False
        >>> with pytest.deprecated_call():
        ...     contains(data, "c", deprecated_keys=["missing", "b"])
        True
    """
    _warnings_hide = True
    if key in obj:
        return True
    for deprecated_key in deprecated_keys:
        if deprecated_key in obj:
            warnings.warn(
                msg.format_map({"deprecated_key": deprecated_key, "key": key}),
                DeprecationWarning,
            )
            return True
    return False


def getitem[KT, VT](
    obj: SupportsGetItem[KT, VT],
    key: KT,
    deprecated_keys: Iterable[KT] = (),
    *,
    msg: str = _DEPRECATED_MESSAGE,
) -> VT:
    """.

    Examples:
        >>> import pytest
        >>> data = {"a": 1, "b": 2}
        >>> getitem(data, "a")
        1
        >>> with pytest.raises(KeyError):
        ...     getitem(data, "missing")
        >>> with pytest.deprecated_call():
        ...     getitem(data, "c", deprecated_keys=["missing", "b"])
        2
    """
    _warnings_hide = True
    try:
        return obj[key]
    except KeyError:
        pass
    for deprecated_key in deprecated_keys:
        try:
            value: VT = obj[deprecated_key]
        except KeyError:
            continue
        warnings.warn(
            msg.format_map({"deprecated_key": deprecated_key, "key": key}),
            DeprecationWarning,
        )
        return value
    raise KeyError(key)



from pathlib import Path

from liblaf.grapes import conf
from liblaf.grapes.conf import BaseConfig, Field


def _default_hide_frame() -> list[str]:
    return ["logging", "rich.progress"]


class ConfigLogging(BaseConfig):
    # ref: <https://pendulum.eustace.io/docs/#tokens>
    datefmt: Field[str] = conf.str(default="YYYY-MM-DD HH:mm:ss.SSS", env="LOG_DATEFMT")
    file: Field[Path | None] = conf.path(default=None, env="LOG_FILE")
    hide_frame: Field[list[str]] = conf.list(factory=_default_hide_frame)
    level: Field[str] = conf.str(default="TRACE")
    time_relative: Field[bool] = conf.bool(default=True, env="LOG_TIME_RELATIVE")



import contextlib
import sys
from typing import Any, Unpack

import wadler_lindig as wl

from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions


def pdoc_array(
    obj: Any, **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc | None:
    size: int | None = _array_size(obj)
    if size is None:
        return None
    if kwargs.get("short_arrays") is None:
        with contextlib.suppress(TypeError):
            kwargs["short_arrays"] = size > kwargs.get("short_arrays_threshold", 100)
    kwargs.pop("custom", None)  # avoid infinite recursion
    return wl.pdoc(obj, **kwargs)


def _array_size(obj: Any) -> int | None:
    for module, type_name in [
        ("jax", "Array"),
        ("mlx.core", "array"),
        ("numpy", "ndarray"),
    ]:
        if module not in sys.modules:
            continue
        typ: type = getattr(sys.modules[module], type_name)
        if isinstance(obj, typ):
            return obj.size
    for module, type_name in [("torch", "Tensor")]:
        if module not in sys.modules:
            continue
        typ: type = getattr(sys.modules[module], type_name)
        if isinstance(obj, typ):
            return obj.numel()
    return None



import math
import statistics
from collections.abc import Sequence

import autoregistry

from liblaf.grapes import pretty

type StatisticName = str


STATISTICS_REGISTRY = autoregistry.Registry(prefix="_compute_")


STATISTICS_REGISTRY["max"] = max
STATISTICS_REGISTRY["mean"] = statistics.mean
STATISTICS_REGISTRY["median"] = statistics.median
STATISTICS_REGISTRY["min"] = min
STATISTICS_REGISTRY["stdev"] = statistics.stdev
STATISTICS_REGISTRY["total"] = sum


def compute_statistic(series: Sequence[float], stat_name: StatisticName) -> float:
    try:
        return STATISTICS_REGISTRY[stat_name](series)
    except (ValueError, statistics.StatisticsError):
        return math.nan


def pretty_statistic(
    series: Sequence[float], stat_name: StatisticName, *, rich_markup: bool = False
) -> tuple[str, str]:
    match stat_name:
        case "mean+stdev":
            mean: float = compute_statistic(series, "mean")
            stdev: float = compute_statistic(series, "stdev")
            pretty_name: str = (
                "[bold green]mean[/bold green] ± [green]σ[/green]"  # noqa: RUF001
                if rich_markup
                else "mean ± σ"  # noqa: RUF001
            )
            mantissas: list[str]
            spacer: str
            units: str
            mantissas, spacer, units = pretty.pretty_durations((mean, stdev))
            pretty_mean: str
            pretty_stdev: str
            pretty_mean, pretty_stdev = mantissas
            if rich_markup:
                pretty_mean = f"[bold green]{pretty_mean}[/bold green]"
                pretty_stdev = f"[green]{pretty_stdev}[/green]"
            if units and rich_markup:
                units = f"[green]{units}[/green]"
            return pretty_name, f"{pretty_mean} ± {pretty_stdev}{spacer}{units}"
        case "range":
            minimum: float = compute_statistic(series, "min")
            maximum: float = compute_statistic(series, "max")
            pretty_name: str = (
                "[cyan]min[/cyan] … [magenta]max[/magenta]"
                if rich_markup
                else "min … max"
            )
            mantissas: list[str]
            spacer: str
            units: str
            mantissas, spacer, units = pretty.pretty_durations((minimum, maximum))
            pretty_min: str
            pretty_max: str
            pretty_min, pretty_max = mantissas
            if rich_markup:
                pretty_min = f"[cyan]{pretty_min}[/cyan]"
                pretty_max = f"[magenta]{pretty_max}[/magenta]"
            if units and rich_markup:
                units = f"[magenta]{units}[/magenta]"
            return pretty_name, f"{pretty_min} … {pretty_max}{spacer}{units}"
        case stat_name:
            value: float = compute_statistic(series, stat_name)
            return stat_name, pretty.pretty_duration(value)



from typing import Any, overload

from liblaf.grapes.sentinel import MISSING


def wrapt_setattr(obj: Any, name: str, value: Any, /) -> None:
    name = f"_self_{name}"
    setattr(obj, name, value)


@overload
def wrapt_getattr(obj: Any, name: str, /) -> Any: ...
@overload
def wrapt_getattr[T](obj: Any, name: str, default: T, /) -> Any | T: ...
def wrapt_getattr(obj: Any, name: str, default: Any = MISSING, /) -> Any:
    name = f"_self_{name}"
    try:
        return getattr(obj, name)
    except AttributeError:
        parent: Any = getattr(obj, "_self_parent", None)
        if parent is None:
            if default is MISSING:
                raise
            return default
        if hasattr(parent, name):
            return getattr(parent, name)
        if default is MISSING:
            raise
        return default



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import attrs

from liblaf.grapes import deps

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure


@attrs.define
class BenchResults:
    outputs: dict[str, list[Any]]
    sizes: list[float]
    timings: dict[str, list[float]]

    def plot(
        self,
        *,
        relative_to: str | None = None,
        xlabel: str = "Size",
        ylabel: str = "Time (sec)",
        log_scale: bool = True,
    ) -> Figure:
        with deps.optional_deps("liblaf-grapes", "bench"):
            import matplotlib.pyplot as plt

        if relative_to is None:
            relative_to = min(self.timings.keys(), key=lambda k: self.timings[k][-1])

        fig: Figure
        ax0: Axes
        ax1: Axes
        fig, (ax0, ax1) = plt.subplots(1, 2, sharex="all", figsize=(12.8, 4.8))

        for label, timings in self.timings.items():
            ax0.plot(self.sizes, timings, label=label)
            base: list[float] = self.timings[relative_to]
            relative: list[float] = [t / b for t, b in zip(timings, base, strict=True)]
            ax1.plot(self.sizes, relative, label=label)

        ax0.grid(which="both", linestyle="--")
        ax0.legend()
        ax0.set_xlabel(xlabel)
        ax0.set_ylabel(ylabel)

        ax1.grid(which="both", linestyle="--")
        ax1.legend()
        ax1.set_xlabel(xlabel)
        ax1.set_ylabel(f"{ylabel} (relative to {relative_to})")

        if log_scale:
            ax0.set_xscale("log")
            ax0.set_yscale("log")
            ax1.set_xscale("log")

        fig.tight_layout()
        return fig



from liblaf.grapes import conf
from liblaf.grapes.conf import BaseConfig

from ._joblib import ConfigJoblib
from ._logging import ConfigLogging
from ._pretty import ConfigPretty
from ._traceback import ConfigTraceback
from ._warnings import ConfigWarnings


class Config(BaseConfig):
    joblib: ConfigJoblib = conf.group()
    logging: ConfigLogging = conf.group()
    pretty: ConfigPretty = conf.group()
    traceback: ConfigTraceback = conf.group()
    warnings: ConfigWarnings = conf.group()


config: Config = Config()



from __future__ import annotations

import logging
from collections.abc import Iterable
from pathlib import Path
from typing import IO, TYPE_CHECKING, override

from rich.console import Console

from liblaf.grapes.rich.logging.handlers import RichHandler, RichHandlerColumn

if TYPE_CHECKING:
    from _typeshed import StrPath


class RichFileHandler(RichHandler):
    def __init__(
        self,
        filename: StrPath,
        mode: str = "w",
        *,
        encoding: str | None = None,
        errors: str | None = None,
        # RichHandler options
        columns: Iterable[RichHandlerColumn] | None = None,
        level: int = logging.NOTSET,
    ) -> None:
        filename = Path(filename)
        filename.parent.mkdir(parents=True, exist_ok=True)
        file: IO[str] = filename.open(mode=mode, encoding=encoding, errors=errors)
        console = Console(file=file)
        super().__init__(console, columns=columns, level=level)

    @override
    def close(self) -> None:
        super().close()
        self.console.file.close()



from ._exception import RichExceptionSummary
from ._frame import RichFrameSummary
from ._install import install
from ._stack import RichStackSummary

__all__ = ["RichExceptionSummary", "RichFrameSummary", "RichStackSummary", "install"]



from collections.abc import Iterable

from ._progress import Progress


def track[T](
    sequence: Iterable[T],
    total: float | None = None,
    completed: int = 0,
    description: str = "Working...",
    *,
    progress: Progress | None = None,
) -> Iterable[T]:
    _logging_hide = True
    if progress is None:
        progress = Progress()
    with progress:
        yield from progress.track(
            sequence, total=total, completed=completed, description=description
        )



import logging
import sys


def remove_non_root_stream_handlers() -> None:
    # loggerDict is not documented, but it's widely used.
    for logger in logging.root.manager.loggerDict.values():
        if not isinstance(logger, logging.Logger):
            continue
        if logger.name == "root":
            continue
        for handler in logger.handlers[:]:  # slice to freeze the list during iteration
            if isinstance(handler, logging.StreamHandler) and (
                handler.stream is sys.stdout or handler.stream is sys.stderr
            ):
                logger.removeHandler(handler)



from collections.abc import Callable
from typing import Any

import wadler_lindig as wl


def chain_custom(
    *custom: Callable[..., wl.AbstractDoc | None] | None,
) -> Callable[[Any], wl.AbstractDoc | None]:
    custom = tuple(func for func in custom if func is not None)
    if not custom:
        return _no_doc
    if len(custom) == 1:
        return custom[0]

    def wrapper(*args, **kwargs) -> wl.AbstractDoc | None:
        for func in custom:
            doc: wl.AbstractDoc | None = func(*args, **kwargs)
            if doc is not None:
                return doc
        return None

    return wrapper


def _no_doc(*_args, **_kwargs) -> None:
    return None



import logging
from collections.abc import Hashable

import attrs


@attrs.define
class OnceFilter:
    _history: set[Hashable] = attrs.field(repr=False, init=False, factory=set)

    def __call__(self, record: logging.LogRecord) -> bool:
        if not getattr(record, "once", False):
            return True
        record_hash: Hashable = self._hash_record(record)
        if record_hash in self._history:
            return False
        self._history.add(record_hash)
        return True

    def _hash_record(self, record: logging.LogRecord) -> Hashable:
        return hash((record.name, record.levelno, record.filename, record.lineno))



# ruff: noqa: A001

import builtins
from decimal import Decimal
from pathlib import Path

from environs import env

from ._field_method import FieldMethod, ListFieldMethod

str: FieldMethod[builtins.str] = FieldMethod(env.str)
bool: FieldMethod[builtins.bool] = FieldMethod(env.bool)
int: FieldMethod[builtins.int] = FieldMethod(env.int)
float: FieldMethod[builtins.float] = FieldMethod(env.float)
decimal: FieldMethod[Decimal] = FieldMethod(env.decimal)
list: ListFieldMethod = ListFieldMethod(env.list)
path: FieldMethod[Path] = FieldMethod(env.path)



from liblaf.grapes import conf
from liblaf.grapes.conf import BaseConfig, Field


class ConfigTraceback(BaseConfig):
    hide_stable_release: Field[bool] = conf.bool(
        default=True, env="TRACEBACK_HIDE_STABLE_RELEASE"
    )
    indent_guide: Field[bool] = conf.bool(default=True)
    locals_hide_dunder: Field[bool] = conf.bool(default=True)
    locals_hide_sunder: Field[bool] = conf.bool(default=True)
    locals_max_length: Field[int] = conf.int(default=10)
    locals_max_string: Field[int] = conf.int(default=80)
    show_locals: Field[bool] = conf.bool(default=True, env="TRACEBACK_SHOW_LOCALS")
    theme: Field[str] = conf.str(default="ansi_dark")



import logging
from collections.abc import Mapping

from liblaf.grapes.icecream import ICECREAM

_DEFAULT_LEVELS: dict[int, str] = {5: "TRACE", ICECREAM: "ICECREAM", 25: "SUCCESS"}


def init_levels(levels: Mapping[int, str] | None = None) -> None:
    if levels is None:
        levels = _DEFAULT_LEVELS
    for level, name in levels.items():
        logging.addLevelName(level, name)



from ._as_iterable import as_iterable
from ._as_sequence import as_sequence
from ._first_not_none import first_not_none
from ._len_or_none import len_or_none
from ._toolz import areidentical, compact, keyjoin, omit, pick

__all__ = [
    "areidentical",
    "as_iterable",
    "as_sequence",
    "compact",
    "first_not_none",
    "keyjoin",
    "len_or_none",
    "omit",
    "pick",
]



from ._autolog import autolog
from ._excepthook import install_excepthook
from ._lazy import LazyRepr, LazyStr
from ._level import init_levels
from ._logger import CleanLogger, set_default_logger_level_by_release_type
from ._remove_handlers import remove_non_root_stream_handlers
from ._tree import LoggerTree
from ._unraisablehook import install_unraisablehook

__all__ = [
    "CleanLogger",
    "LazyRepr",
    "LazyStr",
    "LoggerTree",
    "autolog",
    "init_levels",
    "install_excepthook",
    "install_unraisablehook",
    "remove_non_root_stream_handlers",
    "set_default_logger_level_by_release_type",
]



import functools
from collections.abc import Callable


class LazyRepr[**P]:
    func: Callable[P, str]
    # Pyright doesn't support this annotation
    # args: P.args

    def __init__(
        self, func: Callable[P, str], /, *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __repr__(self) -> str:
        return self.__wrapped__

    @functools.cached_property
    def __wrapped__(self) -> str:
        return self.func(*self.args, **self.kwargs)


class LazyStr[**P]:
    func: Callable[P, str]

    def __init__(
        self, func: Callable[P, str], /, *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return self.__wrapped__

    @functools.cached_property
    def __wrapped__(self) -> str:
        return self.func(*self.args, **self.kwargs)



import inspect
from collections.abc import Callable, Mapping, Sequence
from typing import Any

import attrs
import wadler_lindig as wl

from liblaf.grapes.wadler_lindig import pformat

from ._utils import get_name


@attrs.define
class PrettyCall:
    func_name: str
    args: Sequence[Any] = ()
    kwargs: Mapping[str, Any] = {}

    def __repr__(self) -> str:
        return pformat(self)

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc:
        params: list[wl.AbstractDoc] = [wl.pdoc(arg, **kwargs) for arg in self.args]
        params += wl.named_objs(self.kwargs.items(), **kwargs)
        return wl.bracketed(
            begin=wl.TextDoc(self.func_name) + wl.TextDoc("("),
            docs=params,
            sep=wl.comma,
            end=wl.TextDoc(")"),
            indent=kwargs["indent"],
        )


def pretty_call(
    func: Callable,
    args: Sequence[Any] = (),
    kwargs: Mapping[str, Any] = {},
    **wl_kwargs,
) -> str:
    """.

    Examples:
        >>> def foo(a: int, b: int, c: int = 3): ...
        >>> print(pretty_call(foo, (1, 2), {"c": 4}))
        foo(1, 2, 4)
    """
    func = inspect.unwrap(func)
    args, kwargs = _bind_safe(func, args, kwargs)
    return pformat(PrettyCall(get_name(func), args, kwargs), **wl_kwargs)


def _bind_safe(
    func: Callable, /, args: Sequence[Any], kwargs: Mapping[str, Any]
) -> tuple[Sequence[Any], Mapping[str, Any]]:
    try:
        signature: inspect.Signature = inspect.signature(func)
        arguments: inspect.BoundArguments = signature.bind(*args, **kwargs)
    except TypeError:
        return args, kwargs
    else:
        return arguments.args, arguments.kwargs



import datetime
import functools
from collections.abc import Callable
from typing import Any, Literal, Protocol, TypedDict, overload

import joblib
import wrapt

from liblaf.grapes._config import config
from liblaf.grapes.itertools._toolz import pick

from ._wrapt import wrapt_setattr


class MemorizedFunc[**P, T](Protocol):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...
    @property
    def func(self) -> Callable[P, T]: ...


class Metadata(TypedDict): ...


@overload
def memorize[**P, T](
    func: Callable[P, T],
    /,
    *,
    memory: joblib.Memory | None = ...,
    # memory.cache() params
    ignore: list[str] | None = ...,
    verbose: int | None = ...,
    mmap_mode: Literal["r+", "r", "w+", "c"] | None = ...,
    cache_validation_callback: Callable[[Metadata], bool] | None = ...,
    # memory.reduce_size() params
    bytes_limit: int | str | None = ...,
    items_limit: int | None = ...,
    age_limit: datetime.timedelta | None = ...,
) -> MemorizedFunc[P, T]: ...
@overload
def memorize[**P, T](
    *,
    memory: joblib.Memory | None = None,
    # memory.cache() params
    ignore: list[str] | None = ...,
    verbose: int | None = ...,
    mmap_mode: Literal["r+", "r", "w+", "c"] | None = ...,
    cache_validation_callback: Callable[[Metadata], bool] | None = ...,
    # memory.reduce_size() params
    bytes_limit: int | str | None = ...,
    items_limit: int | None = ...,
    age_limit: datetime.timedelta | None = ...,
) -> Callable[[Callable[P, T]], MemorizedFunc[P, T]]: ...
def memorize(func: Callable | None = None, /, **kwargs: Any) -> Any:
    if func is None:
        return functools.partial(memorize, **kwargs)
    memory: joblib.Memory | None = kwargs.pop("memory", None)
    if memory is None:
        memory = make_memory()
    cache_kwargs: dict[str, Any] = pick(
        {"ignore", "verbose", "mmap_mode", "cache_validation_callback"}, kwargs
    )
    reduce_size_kwargs: dict[str, Any] = pick(
        {"bytes_limit", "items_limit", "age_limit"}, kwargs
    )
    reduce_size_kwargs.setdefault("bytes_limit", config.joblib.memory.bytes_limit.get())

    @wrapt.decorator
    def wrapper[**P, T](
        wrapped: Callable[P, T],
        _instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> T:
        result: Any = wrapped(*args, **kwargs)
        memory.reduce_size(**reduce_size_kwargs)
        return result

    func = memory.cache(func, **cache_kwargs)
    func = wrapper(func)
    wrapt_setattr(func, "memory", memory)
    return func


@functools.cache
def make_memory() -> joblib.Memory:
    return joblib.Memory(location=config.joblib.memory.location.get())



from collections.abc import Mapping
from typing import Any, Unpack

import wadler_lindig as wl

from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions

from ._type import pdoc_type


def pdoc_mapping(
    obj: Mapping[Any, Any], **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc:
    if type(obj) is dict:
        begin: wl.AbstractDoc = wl.TextDoc("{")
    else:
        begin: wl.AbstractDoc = pdoc_type(
            type(obj), dataclass=True, **kwargs
        ) + wl.TextDoc("{")
    return wl.bracketed(
        begin=begin,
        docs=[
            wl.pdoc(key, **kwargs) + wl.TextDoc(": ") + wl.pdoc(value, **kwargs)
            for key, value in obj.items()
        ],
        sep=wl.comma,
        end=wl.TextDoc("}"),
        indent=kwargs.get("indent", 2),
    )



import contextlib
from collections.abc import Generator, Mapping
from typing import Any, Self

import attrs
import tlz
import wadler_lindig as wl
from rich.repr import RichReprResult

from ._constants import METADATA_KEY
from ._entry import Entry
from ._field import Field


class BaseConfigMeta(type):
    def __new__[T: type](
        mcs: type[T],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        /,
        **kwargs: Any,
    ) -> T:
        cls: T = super().__new__(mcs, name, bases, namespace)
        if "__attrs_attrs__" in namespace:
            return cls
        kwargs.setdefault("frozen", True)
        kwargs.setdefault("init", False)
        kwargs.setdefault("repr", False)
        cls = attrs.define(cls, **kwargs)
        return cls


class BaseConfig(metaclass=BaseConfigMeta):
    """.

    Examples:
        >>> from liblaf.grapes import conf
        >>> class Config(BaseConfig):
        ...     a: conf.Field[int] = conf.int(default=0)
        >>> config = Config()
        >>> config.get()
        {'a': 0}
        >>> config.a.get()
        0
        >>> with config.a.overrides(1):
        ...     config.get()
        {'a': 1}
        >>> config.get()
        {'a': 0}
        >>> with config.overrides(a=1):
        ...     config.get()
        {'a': 1}
        >>> config.get()
        {'a': 0}
    """

    def __init__(self, name: str = "") -> None:
        kwargs: dict[str, Any] = {}
        cls: type[BaseConfig] = type(self)
        cls = attrs.resolve_types(cls)
        prefix: str = f"{name}." if name else ""
        for field in attrs.fields(type(self)):
            field: attrs.Attribute
            entry: Entry | None = field.metadata.get(METADATA_KEY, None)
            if entry is None:
                continue
            kwargs[field.name] = entry.make(field, prefix)
        self.__attrs_init__(**kwargs)  # pyright: ignore[reportAttributeAccessIssue]

    def __repr__(self) -> str:
        from liblaf.grapes.wadler_lindig import pformat

        return pformat(self)

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc | None:
        from liblaf.grapes.wadler_lindig import pdoc_rich_repr

        return pdoc_rich_repr(self, **kwargs)

    def __rich_repr__(self) -> RichReprResult:
        from liblaf.grapes.rich.repr import rich_repr_fieldz

        yield from rich_repr_fieldz(self)

    def get(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for f in attrs.fields(type(self)):
            f: attrs.Attribute
            value: Any = getattr(self, f.name)
            if isinstance(value, (BaseConfig, Field)):
                result[f.name] = value.get()
        return result

    def set(self, changes: Mapping[str, Any] = {}, /, **kwargs: Any) -> None:
        changes = tlz.merge(changes, kwargs)
        for key, value in changes.items():
            field: BaseConfig | Field = getattr(self, key)
            field.set(value)

    @contextlib.contextmanager
    def overrides(
        self, changes: Mapping[str, Any] = {}, /, **kwargs: Any
    ) -> Generator[Self]:
        changes = tlz.merge(changes, kwargs)
        with contextlib.ExitStack() as stack:
            for key, value in changes.items():
                field: BaseConfig | Field = getattr(self, key)
                stack.enter_context(field.overrides(value))
            yield self



from __future__ import annotations

from typing import Any

import attrs

from ._config import BaseConfig
from ._constants import METADATA_KEY
from ._entry import Entry


class GroupEntry[T: BaseConfig](Entry[T]):
    def make(self, field: attrs.Attribute, prefix: str) -> T:
        assert field.type is not None
        return field.type(f"{prefix}{field.name}")


def group() -> Any:
    return attrs.field(metadata={METADATA_KEY: GroupEntry()})



import collections
from collections.abc import Callable, Iterable, Sequence

import attrs
import limits

from liblaf.grapes import pretty, warnings
from liblaf.grapes.logging import LazyStr, LimitsHitArgs, autolog

from ._clock import ClockName, clock
from ._statistics import StatisticName, pretty_statistic
from .defaults import (
    DEFAULT_CLOCKS,
    LOG_RECORD_DEFAULT_LEVEL,
    LOG_SUMMARY_DEFAULT_LEVEL,
    LOG_SUMMARY_DEFAULT_STATISTICS,
)


@attrs.define
class Timings:
    label: str | None = attrs.field(default=None)
    name: str | None = attrs.field(default=None)
    clocks: Sequence[ClockName] = attrs.field(default=DEFAULT_CLOCKS)
    timings: dict[ClockName, list[float]] = attrs.field(
        factory=lambda: collections.defaultdict(list), init=False
    )

    _start_time: dict[ClockName, float] = attrs.field(factory=dict, init=False)
    _stop_time: dict[ClockName, float] = attrs.field(factory=dict, init=False)

    def __attrs_post_init__(self) -> None:
        _warnings_hide = True
        if self.label is None and self.name is not None:
            warnings.warn(
                "'name' parameter is deprecated. Please use 'label' instead.",
                DeprecationWarning,
            )
            self.label = self.name

    def __len__(self) -> int:
        return len(self.timings[self.default_clock])

    @property
    def default_clock(self) -> ClockName:
        return self.clocks[0]

    def clear(self) -> None:
        self.timings.clear()
        self._start_time.clear()
        self._stop_time.clear()

    def elapsed(self, clock_name: ClockName | None = None) -> float:
        clock_name = clock_name or self.default_clock
        stop_time: float
        if clock_name in self._stop_time:
            stop_time = self._stop_time[clock_name]
        else:
            stop_time = clock(clock_name)
        return stop_time - self._start_time[clock_name]

    def log_record(
        self,
        *,
        index: int = -1,
        level: int = LOG_RECORD_DEFAULT_LEVEL,
        limits: str | limits.RateLimitItem | None = "1/second",
    ) -> None:
        _logging_hide = True
        autolog.log(
            level,
            "%s",
            LazyStr(lambda: self.pretty_record(index=index)),
            extra={
                "limits": LimitsHitArgs(
                    item=limits, identifiers=(self.label or "Timer",)
                )
            },
        )

    def log_summary(
        self,
        *,
        level: int = LOG_SUMMARY_DEFAULT_LEVEL,
        stats: Iterable[StatisticName] = LOG_SUMMARY_DEFAULT_STATISTICS,
        limits: str | limits.RateLimitItem | None = None,
    ) -> None:
        _logging_hide = True
        autolog.log(
            level,
            "%s",
            LazyStr(lambda: self.pretty_summary(stats=stats)),
            extra={
                "limits": LimitsHitArgs(
                    item=limits, identifiers=(self.label or "Timer",)
                ),
                "markup": lambda: self.pretty_summary(stats=stats, rich_markup=True),
            },
        )

    def pretty_record(self, index: int = -1) -> str:
        name: str = self.label or "Timer"
        items: list[str] = [
            f"{clock_name}: {pretty.pretty_duration(self.timings[clock_name][index])}"
            for clock_name in self.clocks
        ]
        items_str: str = ", ".join(items)
        return f"{name} > {items_str}"

    def pretty_summary(
        self,
        stats: Iterable[StatisticName] = LOG_SUMMARY_DEFAULT_STATISTICS,
        *,
        rich_markup: bool = False,
    ) -> str:
        name: str = self.label or "Timer"
        header: str = f"{name} (count: {len(self)})"
        if len(self) == 0:
            return header
        lines: list[str] = []
        for clock_name in self.clocks:
            stats_str: list[str] = []
            for stat in stats:
                stat_name: str
                value: str
                stat_name, value = pretty_statistic(
                    self.timings[clock_name], stat, rich_markup=rich_markup
                )
                stats_str.append(f"{stat_name}: {value}")
            line: str = f"{clock_name} > {', '.join(stats_str)}"
            lines.append(line)
        if len(self.clocks) == 1:
            return f"{header} {lines[0]}"
        return f"{header}\n" + "\n".join(lines)


type Callback = Callable[[Timings], None]



import functools
from collections.abc import Callable
from typing import Any, overload

from liblaf.grapes.fieldz import has_fields

from ._fieldz import rich_repr_fieldz


@overload
def auto_rich_repr[T: type](cls: T, *, rich_repr: bool | None = None) -> T: ...
@overload
def auto_rich_repr[T: type](*, rich_repr: bool | None = None) -> Callable[[T], T]: ...
def auto_rich_repr(cls: type | None = None, *, rich_repr: bool | None = None) -> Any:
    if cls is None:
        return functools.partial(auto_rich_repr, rich_repr=rich_repr)
    if rich_repr is None:
        rich_repr = not hasattr(cls, "__rich_repr__")
    if rich_repr and has_fields(cls):
        cls.__rich_repr__ = rich_repr_fieldz
    return cls



import functools
import logging
import math
import multiprocessing
import time
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Any, overload

import attrs

from ._results import BenchResults

logger: logging.Logger = logging.getLogger(__name__)


def _default_setup() -> Iterable[tuple[Sequence, Mapping]]:
    yield (), {}


def _default_size_fn(*_args, **_kwargs) -> int:
    return 1


@attrs.define
class Bencher:
    min_time: float = 0.2
    timeout: float = 10.0
    warmup: int = 1
    _registry: dict[str, Callable] = attrs.field(init=False, factory=dict)
    _setup: Callable[..., Iterable[tuple[Sequence, Mapping]]] = attrs.field(
        default=_default_setup, init=False
    )
    _size_fn: Callable[..., float] = attrs.field(default=_default_size_fn, init=False)

    @overload
    def bench[C: Callable](self, func: C, /, *, label: str | None = None) -> C: ...
    @overload
    def bench[C: Callable](self, *, label: str | None = None) -> Callable[[C], C]: ...
    def bench(
        self, func: Callable | None = None, *, label: str | None = None
    ) -> Callable:
        if func is None:
            return functools.partial(self.bench, label=label)
        if label is None:
            label = func.__name__
        self._registry[label] = func
        return func

    def setup[C: Callable[..., Iterable[tuple[Sequence, Mapping]]]](self, func: C) -> C:
        self._setup = func
        return func

    def size[C: Callable[..., float]](self, func: C) -> C:
        self._size_fn = func
        return func

    def run(self) -> BenchResults:
        inputs: Sequence[tuple[Sequence, Mapping]] = list(self._setup())
        sizes_and_inputs: list[tuple[float, tuple[Sequence, Mapping]]] = [
            (self._size_fn(*args, **kwargs), (args, kwargs)) for args, kwargs in inputs
        ]
        sizes_and_inputs.sort(key=lambda x: x[0])
        sizes: Sequence[float] = [size for size, _ in sizes_and_inputs]
        inputs = [(args, kwargs) for _, (args, kwargs) in sizes_and_inputs]
        outputs: dict[str, list[Any]] = {}
        timings: dict[str, list[float]] = {}
        for label, func in self._registry.items():
            outputs_name: list[Any] = []
            timings_name: list[float] = []
            for size, (args, kwargs) in zip(sizes, inputs, strict=True):
                output: Any
                elapsed: float
                output, elapsed = self._bench(func, args, kwargs)
                outputs_name.append(output)
                timings_name.append(elapsed)
                logger.debug("Bench %s(%g) took %g sec", label, size, elapsed)
                if elapsed > self.timeout:
                    logger.warning("Bench %s(%g) timed out", label, size)
                    break
            if len(timings_name) < len(sizes):
                n_remain: int = len(sizes) - len(timings_name)
                outputs_name.extend([None] * n_remain)
                timings_name.extend([math.inf] * n_remain)
            timings[label] = timings_name
            outputs[label] = outputs_name
        return BenchResults(outputs=outputs, sizes=list(sizes), timings=timings)

    def _bench[T](
        self, func: Callable[..., T], args: Sequence, kwargs: Mapping
    ) -> tuple[T | None, float]:
        result: T | None = None
        for _ in range(self.warmup):
            result = func(*args, **kwargs)
        count: int = 0
        total_elapsed: float = 0.0
        while total_elapsed < self.min_time:
            start: float = time.perf_counter()
            result = func(*args, **kwargs)
            end: float = time.perf_counter()
            count += 1
            total_elapsed += end - start
        return result, total_elapsed / count

    def _bench_process[T](
        self, func: Callable[..., T], args: Sequence, kwargs: Mapping
    ) -> tuple[T | None, float]:
        def wrapper(
            queue: multiprocessing.Queue, *args: Sequence, **kwargs: Mapping
        ) -> None:
            result: T | None = None
            for _ in range(self.warmup):
                result = func(*args, **kwargs)
            count: int = 0
            total_elapsed: float = 0.0
            while total_elapsed < self.min_time:
                start: float = time.perf_counter()
                result = func(*args, **kwargs)
                end: float = time.perf_counter()
                count += 1
                total_elapsed += end - start
            queue.put((result, total_elapsed / count))

        queue: multiprocessing.Queue[tuple[T, float]] = multiprocessing.Queue()
        process = multiprocessing.Process(
            target=wrapper, args=(queue, *args), kwargs=kwargs
        )
        process.start()
        process.join(self.timeout)
        if process.is_alive():
            process.kill()
            return None, math.inf
        result: T
        elapsed: float
        result, elapsed = queue.get()
        return result, elapsed



import inspect
import types
from collections.abc import Callable, Iterable, Mapping

from liblaf.grapes._config import config

from ._release_type import is_pre_release


def hidden_from_logging(frame: types.FrameType | None) -> bool:
    if frame is None:
        return False
    # `__logging_hide` does not work as expected in some cases due to name mangling
    # `__logging_hide__` will violate Ruff F841
    # so we use `_logging_hide` here
    if _coalesce(frame.f_locals, ("_logging_hide", "__tracebackhide__")):
        return True
    name: str = frame.f_globals.get("__name__", "")
    return f"{name}.".startswith(tuple(config.logging.hide_frame.get()))


def hidden_from_traceback(
    frame: types.FrameType | None, *, hide_stable_release: bool | None = None
) -> bool:
    if frame is None:
        return False
    if _coalesce(frame.f_locals, ("__tracebackhide__",)):
        return True
    if hide_stable_release is None:
        hide_stable_release = config.traceback.hide_stable_release.get()
    if hide_stable_release:
        name: str | None = frame.f_globals.get("__name__")
        if not is_pre_release(frame.f_code.co_filename, name):
            return True
    return False


def hidden_from_warnings(
    frame: types.FrameType | None, *, hide_stable_release: bool | None = None
) -> bool:
    if frame is None:
        return False
    if _coalesce(frame.f_locals, ("_warnings_hide", "__tracebackhide__")):
        return True
    if hide_stable_release is None:
        hide_stable_release = config.traceback.hide_stable_release.get()
    if hide_stable_release:
        name: str | None = frame.f_globals.get("__name__")
        if not is_pre_release(frame.f_code.co_filename, name):
            return True
    return False


def get_frame(
    depth: int = 1, hidden: Callable[[types.FrameType], bool] | None = None
) -> types.FrameType | None:
    frame: types.FrameType | None = inspect.currentframe()
    while frame is not None and depth > 0:
        frame = frame.f_back
        depth -= 1
        if hidden is not None:
            while frame is not None and hidden(frame):
                frame = frame.f_back
    return frame


def get_frame_with_stacklevel(
    depth: int = 1, hidden: Callable[[types.FrameType], bool] | None = None
) -> tuple[types.FrameType | None, int]:
    frame: types.FrameType | None = inspect.currentframe()
    stacklevel: int = 0
    while frame is not None and depth > 0:
        frame = frame.f_back
        depth -= 1
        stacklevel += 1
        if hidden is not None:
            while frame is not None and hidden(frame):
                frame = frame.f_back
                stacklevel += 1
    return frame, stacklevel


def _coalesce[KT, VT](
    obj: Mapping[KT, VT], keys: Iterable[KT], *, default: bool = False
) -> bool:
    for key in keys:
        try:
            return bool(obj[key])
        except (KeyError, ValueError):
            continue
    return default



def first_not_none[T](*args: T | None) -> T:
    """Returns the first `not None` value in the `args`.

    Examples:
        >>> first_not_none(1, 2)
        1
        >>> first_not_none(None, 1)
        1

    References:
        1. [`more_itertools.first_true`](https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.first_true)
    """
    return next(arg for arg in args if arg is not None)



import os
import time
from typing import Literal

import autoregistry

# for code-completion
type ClockName = Literal[
    "monotonic",
    "perf",
    "process",
    "thread",
    "time",
    "children-system",
    "children-user",
    "elapsed",
    "system",
    "user",
]

CLOCK_REGISTRY = autoregistry.Registry()

CLOCK_REGISTRY["monotonic"] = time.monotonic
CLOCK_REGISTRY["perf"] = time.perf_counter
CLOCK_REGISTRY["process"] = time.process_time
CLOCK_REGISTRY["thread"] = time.thread_time
CLOCK_REGISTRY["time"] = time.time

CLOCK_REGISTRY["children-system"] = lambda: os.times().children_system
CLOCK_REGISTRY["children-user"] = lambda: os.times().children_user
CLOCK_REGISTRY["elapsed"] = lambda: os.times().elapsed
CLOCK_REGISTRY["system"] = lambda: os.times().system
CLOCK_REGISTRY["user"] = lambda: os.times().user


def clock(name: ClockName = "perf") -> float:
    return CLOCK_REGISTRY[name]()



import logging
from typing import Any, override

import limits
from rich.console import Console
from rich.progress import (
    BarColumn,
    GetTimeCallable,
    MofNCompleteColumn,
    ProgressColumn,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.progress import Progress as RichProgress

from liblaf.grapes.logging.helpers import autolog

from ._rate_column import RateColumn


class Progress(RichProgress):
    limit: limits.RateLimitItem
    limiter: limits.strategies.RateLimiter
    logger: logging.Logger
    stale: bool = False

    def __init__(
        self,
        *columns: str | ProgressColumn,
        limit: str | limits.RateLimitItem | None = None,
        limiter: limits.strategies.RateLimiter | None = None,
        logger: logging.Logger | None = None,
        # RichProgress options
        speed_estimate_period: float = 30.0,
        get_time: GetTimeCallable | None = None,
        disable: bool = False,
        expand: bool = False,
    ) -> None:
        super().__init__(
            *columns,
            console=Console(quiet=True),
            auto_refresh=False,
            speed_estimate_period=speed_estimate_period,
            get_time=get_time,
            disable=disable,
            expand=expand,
        )
        if limit is None:
            limit = limits.RateLimitItemPerSecond(1)
        elif isinstance(limit, str):
            limit = limits.parse(limit)
        self.limit = limit
        if limiter is None:
            limiter = limits.strategies.SlidingWindowCounterRateLimiter(
                limits.storage.MemoryStorage()
            )
        self.limiter = limiter
        if logger is None:
            logger = autolog
        self.logger = logger

    @override
    @classmethod
    def get_default_columns(cls) -> tuple[str | ProgressColumn, ...]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return (
            # SpinnerColumn(),
            TextColumn("{task.description}", style="progress.description"),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            "[",
            TimeElapsedColumn(),
            "<",
            TimeRemainingColumn(),
            ",",
            RateColumn(),
            "]",
        )

    @override
    def advance(self, task_id: TaskID, advance: float = 1) -> None:
        _logging_hide = True
        self.stale = True
        super().advance(task_id, advance)
        self.refresh()

    @override
    def refresh(self, *, force: bool = False) -> None:
        _logging_hide = True
        if self.stale and (self.limiter.hit(self.limit) or force):
            self.stale = False
            self.logger.info(self.get_renderable())

    @override
    def reset(
        self,
        task_id: TaskID,
        *,
        start: bool = True,
        total: float | None = None,
        completed: int = 0,
        visible: bool | None = None,
        description: str | None = None,
        **fields: Any,
    ) -> None:
        _logging_hide = True
        self.stale = True
        super().reset(
            task_id,
            start=start,
            total=total,
            completed=completed,
            visible=visible,
            description=description,
            **fields,
        )

    @override
    def start(self) -> None:
        pass

    @override
    def stop(self) -> None:
        _logging_hide = True
        self.refresh(force=True)

    @override
    def update(
        self,
        task_id: TaskID,
        *,
        total: float | None = None,
        completed: float | None = None,
        advance: float | None = None,
        description: str | None = None,
        visible: bool | None = None,
        refresh: bool = False,
        **fields: Any,
    ) -> None:
        _logging_hide = True
        self.stale = True
        super().update(
            task_id,
            total=total,
            completed=completed,
            advance=advance,
            description=description,
            visible=visible,
            refresh=refresh,
            **fields,
        )



import ast
import logging
import textwrap
import types
from collections.abc import Mapping, Sequence
from typing import Any, Unpack

import attrs
import wadler_lindig as wl
from asttokens import ASTTokens

from liblaf.grapes import magic
from liblaf.grapes.wadler_lindig import WadlerLindigOptions, pdoc

from ._source import Source

ICON: str = "🍦"
ICECREAM: int = 15  # between DEBUG (10) and INFO (20)


@attrs.define
class IceCreamDebugger:
    enabled: bool = True

    def __call__(self, *args, **kwargs) -> Any:
        if self.enabled:
            frame: types.FrameType | None = magic.get_frame(depth=2)
            logger: logging.Logger = self._get_logger(frame)
            logger.log(ICECREAM, self._format(args, kwargs, frame), stacklevel=2)
        match len(args):
            case 0:
                return None
            case 1:
                return args[0]
            case _:
                return args

    def _annotate_args(
        self, args: Sequence[Any], frame: types.FrameType | None = None
    ) -> list[tuple[str | None, Any]]:
        call: ast.Call | None = None
        if frame is not None:
            call = Source.executing(frame).node  # pyright: ignore[reportAssignmentType]
        if frame is None or call is None:
            return [(None, arg) for arg in args]
        source: Source = Source.for_frame(frame)
        asttokens: ASTTokens = source.asttokens()
        result: list[tuple[str | None, Any]] = []
        for value, node in zip(args, call.args, strict=True):
            text: str = asttokens.get_text(node)
            if _is_literal(text):
                result.append((None, value))
                continue
            if "\n" in text:
                text = " " * node.col_offset + text
                text = textwrap.dedent(text)
            text = text.strip()
            result.append((text, value))
        return result

    def _format(
        self,
        args: Sequence[Any],
        kwargs: Mapping[str, Any] = {},
        frame: types.FrameType | None = None,
    ) -> str:
        if len(args) == 0:
            if frame is not None:
                context: str = self._format_context(frame)
                return f"{ICON} {context}"
            return ICON
        pairs: list[tuple[str | None, Any]] = self._annotate_args(args, frame)
        doc: wl.AbstractDoc = self._pdoc_args(pairs, **kwargs)
        return wl.pformat(doc)

    def _format_context(self, frame: types.FrameType) -> str:
        filename: str = frame.f_code.co_filename
        filename = magic.abbr_path(filename)
        lineno: int = frame.f_lineno
        name: str = frame.f_code.co_name
        if name != "<module>":
            name += "()"
        return f"{filename}:{lineno} in {name}"

    def _get_logger(self, frame: types.FrameType | None) -> logging.Logger:
        _logging_hide = True
        name: str | None = None
        if frame is not None:
            name = frame.f_globals.get("__name__")
        if name is None:
            name = "icecream"
        return logging.getLogger(name)

    def _pdoc_args(
        self,
        args: Sequence[tuple[str | None, Any]],
        **kwargs: Unpack[WadlerLindigOptions],
    ) -> wl.AbstractDoc:
        docs: list[wl.AbstractDoc] = []
        for name, value in args:
            value_doc: wl.AbstractDoc = pdoc(value, **kwargs)
            arg_doc: wl.AbstractDoc
            if name is None:
                arg_doc = value_doc
            else:
                arg_doc = wl.TextDoc(name) + wl.TextDoc(": ") + value_doc
            docs.append(arg_doc)
        return wl.join(wl.BreakDoc(", "), docs)


ic = IceCreamDebugger()


def _is_literal(s: str) -> bool:
    try:
        ast.literal_eval(s)
    except Exception:  # noqa: BLE001
        return False
    else:
        return True



import functools
from collections.abc import Callable
from typing import Any, overload

from liblaf.grapes.fieldz import has_fields

from ._pformat import pformat
from .custom import pdoc_fieldz, pdoc_rich_repr


@overload
def auto_pdoc[T: type](
    cls: T, *, pdoc: bool | None = None, repr: bool | None = None
) -> T: ...
@overload
def auto_pdoc[T: type](
    *, pdoc: bool | None = None, repr: bool | None = None
) -> Callable[[T], T]: ...
def auto_pdoc(
    cls: type | None = None,
    *,
    pdoc: bool | None = None,
    repr: bool | None = None,  # noqa: A002
) -> Any:
    if cls is None:
        return functools.partial(auto_pdoc, pdoc=pdoc, repr=repr)
    if pdoc is None:
        pdoc = not hasattr(cls, "__pdoc__")
    if repr is None:
        repr = cls.__repr__ is object.__repr__  # noqa: A001
    if pdoc:
        if hasattr(cls, "__rich_repr__"):
            cls.__pdoc__ = pdoc_rich_repr
        elif has_fields(cls):
            cls.__pdoc__ = pdoc_fieldz
    if repr:
        cls.__repr__ = pformat  # pyright: ignore[reportAttributeAccessIssue]
    return cls



from . import filters, handlers, helpers
from ._init import init
from .filters import LimitsFilter, LimitsHitArgs
from .handlers import RichFileHandler
from .helpers import (
    CleanLogger,
    LazyRepr,
    LazyStr,
    LoggerTree,
    autolog,
    init_levels,
    install_excepthook,
    install_unraisablehook,
    remove_non_root_stream_handlers,
    set_default_logger_level_by_release_type,
)

__all__ = [
    "CleanLogger",
    "LazyRepr",
    "LazyStr",
    "LimitsFilter",
    "LimitsHitArgs",
    "LoggerTree",
    "RichFileHandler",
    "autolog",
    "filters",
    "handlers",
    "helpers",
    "init",
    "init_levels",
    "install_excepthook",
    "install_unraisablehook",
    "remove_non_root_stream_handlers",
    "set_default_logger_level_by_release_type",
]



from . import custom
from ._auto import auto_pdoc
from ._options import make_kwargs
from ._pdoc import pdoc
from ._pformat import pformat
from ._pprint import pprint
from ._typing import CustomCallable, WadlerLindigOptions
from .custom import (
    PdocCustomDispatcher,
    chain_custom,
    pdoc_array,
    pdoc_custom,
    pdoc_datetime,
    pdoc_enum,
    pdoc_fieldz,
    pdoc_mapping,
    pdoc_pydantic_root_model,
    pdoc_rich_repr,
    pdoc_type,
)

__all__ = [
    "CustomCallable",
    "PdocCustomDispatcher",
    "WadlerLindigOptions",
    "auto_pdoc",
    "chain_custom",
    "custom",
    "make_kwargs",
    "pdoc",
    "pdoc_array",
    "pdoc_custom",
    "pdoc_datetime",
    "pdoc_enum",
    "pdoc_fieldz",
    "pdoc_mapping",
    "pdoc_pydantic_root_model",
    "pdoc_rich_repr",
    "pdoc_type",
    "pformat",
    "pprint",
]



import logging
from collections.abc import Iterable, Sequence

from ._clock import ClockName
from ._statistics import StatisticName

DEFAULT_CLOCKS: Sequence[ClockName] = ("perf",)


LOG_RECORD_DEFAULT_LEVEL: int = logging.DEBUG


LOG_SUMMARY_DEFAULT_LEVEL: int = logging.INFO
LOG_SUMMARY_DEFAULT_STATISTICS: Iterable[StatisticName] = (
    "total",
    "mean+stdev",
    "range",
    # "median",
)



from typing import Any, Unpack

import wadler_lindig as wl

from ._options import WadlerLindigOptions, make_kwargs


def pprint(obj: Any, **kwargs: Unpack[WadlerLindigOptions]) -> None:
    """Pretty-prints an object to stdout.

    Examples:
        array:

        >>> import numpy as np
        >>> pprint(np.zeros((2, 3)))
        array([[0., 0., 0.],
               [0., 0., 0.]])
        >>> pprint(np.zeros((20, 30)))
        f64[20,30](numpy)

        datetime:

        >>> import datetime
        >>> pprint(datetime.datetime(1970, 1, 1, tzinfo=datetime.UTC))
        1970-01-01T00:00:00+00:00

        dataclass:

        >>> import dataclasses
        >>> import numpy as np
        >>> @dataclasses.dataclass
        ... class MyDataclass:
        ...     x: list[str]
        ...     y: np.ndarray
        >>> obj = MyDataclass(["lorem", "ipsum", "dolor sit amet"], np.zeros((2, 3)))
        >>> pprint(obj, width=30, indent=4)
        MyDataclass(
            x=[
                'lorem',
                'ipsum',
                'dolor sit amet'
            ],
            y=array([[0., 0., 0.],
                     [0., 0., 0.]])
        )

        dict:

        >>> pprint({"jack": 4098, "sjoerd": 4127})
        {'jack': 4098, 'sjoerd': 4127}

        pydantic:

        >>> import pydantic
        >>> class MyModel(pydantic.RootModel[list[str]]): ...
        >>> obj = MyModel(["lorem", "ipsum", "dolor sit amet"])
        >>> pprint(obj)
        MyModel['lorem', 'ipsum', 'dolor sit amet']

        rich.repr:

        >>> import rich
        >>> class Bird:
        ...     def __rich_repr__(self):
        ...         yield "penguin"
        ...         yield "eats", ["fish"]
        ...         yield "fly", False, True
        ...         yield "extinct", False, False
        >>> pprint(Bird())
        Bird('penguin', eats=['fish'], fly=False)
    """
    kwargs: WadlerLindigOptions = make_kwargs(**kwargs)
    return wl.pprint(obj, **kwargs)



import contextlib
import contextvars
from collections.abc import Callable, Generator
from typing import cast

import attrs
import environs
import wadler_lindig as wl
from rich.repr import RichReprResult

from liblaf.grapes.sentinel import MISSING, MissingType


@attrs.frozen
class Field[T]:
    """.

    Examples:
        >>> from environs import env
        >>> a = Field("a", default=0, getter=env.int)
        >>> a
        Field(name='a', value=0)
        >>> a.name
        'a'
        >>> a.get()
        0
        >>> with a.overrides(1):
        ...     a.get()
        1
        >>> a.get()
        0
        >>> a = Field("a", factory=lambda: 0, getter=env.int)
        >>> a.get()
        0
    """

    _var: contextvars.ContextVar[T] = attrs.field(repr=False, alias="var")

    def __init__(
        self,
        name: str,
        *,
        default: T | MissingType = MISSING,
        env: str | None = None,
        factory: Callable[[], T] | None = None,
        getter: Callable[[str], T],
    ) -> None:
        value: T | MissingType = _get_value(
            name, default=default, env=env, factory=factory, getter=getter
        )
        var: contextvars.ContextVar[T]
        if value is MISSING:
            var = contextvars.ContextVar(name)
        else:
            value = cast("T", value)
            var = contextvars.ContextVar(name, default=value)
        self.__attrs_init__(var=var)  # pyright: ignore[reportAttributeAccessIssue]

    def __repr__(self) -> str:
        from liblaf.grapes.wadler_lindig import pformat

        return pformat(self)

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc | None:
        from liblaf.grapes.wadler_lindig import pdoc_rich_repr

        return pdoc_rich_repr(self, **kwargs)

    def __rich_repr__(self) -> RichReprResult:
        yield "name", self.name
        yield "value", self.get()

    @property
    def name(self) -> str:
        return self._var.name

    def get(self) -> T:
        return self._var.get()

    def set(self, value: T, /) -> None:
        self._var.set(value)

    @contextlib.contextmanager
    def overrides(self, value: T, /) -> Generator[T]:
        token: contextvars.Token[T] = self._var.set(value)
        try:
            yield value
        finally:
            self._var.reset(token)


def _get_value[T](
    name: str,
    *,
    default: T | MissingType = MISSING,
    env: str | None = None,
    factory: Callable[[], T] | None = None,
    getter: Callable[[str], T],
) -> T | MissingType:
    if env is None:
        env = name.upper().replace(".", "_")
    try:
        return getter(env)
    except environs.EnvError:
        pass
    if default is not MISSING:
        return default
    if factory is not None:
        return factory()
    return MISSING



# ref: <https://toolz.readthedocs.io/en/latest/tips-and-tricks.html>

import itertools
from collections.abc import Container, Iterable, Iterator, Mapping
from typing import Any

import tlz


def pick[KT, VT](allowlist: Container[KT], dictionary: Mapping[KT, VT]) -> dict[KT, VT]:
    """Return a subset of the provided dictionary with keys contained in the allowlist.

    Examples:
        >>> alphabet = {"a": 1, "b": 2, "c": 3, "d": 4}
        >>> pick(["a", "b"], alphabet)
        {'a': 1, 'b': 2}
    """
    return tlz.keyfilter(lambda k: k in allowlist, dictionary)


def omit[KT, VT](denylist: Container[KT], dictionary: Mapping[KT, VT]) -> dict[KT, VT]:
    """Return a subset of the provided dictionary with keys not contained in the denylist.

    Examples:
        >>> alphabet = {"a": 1, "b": 2, "c": 3, "d": 4}
        >>> omit(["a", "b"], alphabet)
        {'c': 3, 'd': 4}
    """
    return tlz.keyfilter(lambda k: k not in denylist, dictionary)


def compact[T](iterable: Iterable[T | None]) -> Iterator[T]:
    """Filter an iterable on "truthy" values.

    Examples:
        >>> results = [0, 1, 2, None, 3, False]
        >>> list(compact(results))
        [1, 2, 3]
    """
    return tlz.filter(None, iterable)


def keyjoin[KT, VT](
    leftkey: KT,
    leftseq: Iterable[Mapping[KT, VT]],
    rightkey: KT,
    rightseq: Iterable[Mapping[KT, VT]],
) -> Iterator[Mapping[KT, VT]]:
    """Inner join two sequences of dictionaries on specified keys, merging matches with right value precedence.

    Examples:
        >>> people = [
        ...     {"id": 0, "name": "Anonymous Guy", "location": "Unknown"},
        ...     {"id": 1, "name": "Karan", "location": "San Francisco"},
        ...     {"id": 2, "name": "Matthew", "location": "Oakland"},
        ... ]
        >>> hobbies = [
        ...     {"person_id": 1, "hobby": "Tennis"},
        ...     {"person_id": 1, "hobby": "Acting"},
        ...     {"person_id": 2, "hobby": "Biking"},
        ... ]
        >>> list(keyjoin("id", people, "person_id", hobbies))
        [{'id': 1, 'name': 'Karan', 'location': 'San Francisco', 'person_id': 1, 'hobby': 'Tennis'}, {'id': 1, 'name': 'Karan', 'location': 'San Francisco', 'person_id': 1, 'hobby': 'Acting'}, {'id': 2, 'name': 'Matthew', 'location': 'Oakland', 'person_id': 2, 'hobby': 'Biking'}]
    """
    return itertools.starmap(tlz.merge, tlz.join(leftkey, leftseq, rightkey, rightseq))


def areidentical(*seqs: Iterable[Any]) -> bool:
    """Determine if sequences are identical element-wise. This lazily evaluates the sequences and stops as soon as the result is determined.

    Examples:
        >>> areidentical([1, 2, 3], (1, 2, 3))
        True

        >>> areidentical([1, 2, 3], [1, 2])
        False
    """
    return not any(tlz.diff(*seqs, default=object()))



from ._memorize import MemorizedFunc, memorize
from ._wraps import wraps
from ._wrapt import wrapt_getattr, wrapt_setattr

__all__ = ["MemorizedFunc", "memorize", "wraps", "wrapt_getattr", "wrapt_setattr"]



import logging
import sys
import types
from typing import ClassVar, override

from liblaf.grapes import magic


class CleanLogger(logging.Logger):
    dev_level: ClassVar[int | str] = 1
    pre_level: ClassVar[int | str] = logging.DEBUG

    def __init__(self, name: str, level: int | str = logging.NOTSET) -> None:
        _logging_hide = True
        super().__init__(name, level)
        if level != logging.NOTSET:
            return
        module: types.ModuleType | None = sys.modules.get(name)
        file: str | None = None
        if module is not None:
            file = getattr(module, "__file__", None)
        if file is None:
            frame: types.FrameType | None = magic.get_frame(
                hidden=magic.hidden_from_logging
            )
            if frame is not None:
                file = frame.f_code.co_filename
        if magic.is_dev_release(file, name):
            self.setLevel(self.dev_level)
        elif magic.is_pre_release(file, name):
            self.setLevel(self.pre_level)

    @property
    def propagate(self) -> bool:  # pyright: ignore[reportIncompatibleVariableOverride]
        return True

    @propagate.setter
    def propagate(self, value: bool) -> None:  # pyright: ignore[reportIncompatibleVariableOverride]
        pass

    @override
    def addHandler(self, hdlr: logging.Handler) -> None:
        if self.name != "root" and isinstance(hdlr, logging.StreamHandler):
            return
        super().addHandler(hdlr)


def set_default_logger_level_by_release_type(
    dev_level: int | str | None = None, pre_level: int | str | None = None
) -> None:
    if dev_level is not None:
        CleanLogger.dev_level = dev_level
    if pre_level is not None:
        CleanLogger.pre_level = pre_level
    logging.setLoggerClass(CleanLogger)



import logging
from typing import override

import attrs
from rich.align import AlignMethod
from rich.text import Text

from ._abc import RichHandlerColumn

_LEVEL_ALIASES: dict[int, dict[str, str]] = {
    1: {"ICECREAM": "M"},
    3: {
        "NOTSET": "NST",
        "TRACE": "TRC",
        "DEBUG": "DBG",
        "ICECREAM": "ICM",
        "INFO": "INF",
        "SUCCESS": "SUC",
        "WARNING": "WRN",
        "ERROR": "ERR",
        "CRITICAL": "CRT",
    },
}


@attrs.define
class RichHandlerColumnLevel(RichHandlerColumn):
    align: AlignMethod = "right"
    width: int = 3

    @override
    def render(self, record: logging.LogRecord) -> Text:
        level: str = self._get_abbr(record)
        text = Text(level, f"logging.level.{record.levelname.lower()}")
        text.align(self.align, self.width)
        return text

    def _get_abbr(self, record: logging.LogRecord) -> str:
        if record.levelname == f"Level {record.levelno}":
            return str(record.levelno)
        aliases: dict[str, str] = _LEVEL_ALIASES.get(self.width, {})
        return aliases.get(record.levelname, record.levelname)



import collections
import logging
from collections.abc import Mapping

from rich.panel import Panel
from rich.pretty import Pretty
from rich.tree import Tree


class LoggerTree(collections.UserDict[str, Tree]):
    def __init__(
        self, mapping: Mapping[str, Tree] | None = None, /, **kwargs: Tree
    ) -> None:
        super().__init__(mapping, **kwargs)
        for logger in logging.root.manager.loggerDict.values():
            if isinstance(logger, logging.Logger):
                self.add_logger(logger)

    def __rich__(self) -> Tree:
        return self["root"]

    def add_logger(self, logger: logging.Logger) -> Tree:
        if (tree := self.get(logger.name)) is not None:
            return tree
        tree = self._rich_logger(logger)
        self.data[logger.name] = tree
        if (parent := logger.parent) is not None:
            parent_tree: Tree = self.add_logger(parent)
            parent_tree.add(tree)
        return tree

    def _rich_logger(self, logger: logging.Logger) -> Tree:
        tree = Tree(Pretty(logger))
        for filter_ in logger.filters:
            tree.add(Pretty(filter_))
        for handler in logger.handlers:
            handler_node: Tree = tree.add(Pretty(handler))
            for filter_ in handler.filters:
                handler_node.add(Pretty(filter_))
        if len(tree.children) > 0:
            tree = Tree(Panel(tree))
        return tree



from typing import overload

ICECREAM: int
ICON: str

class IceCreamDebugger:
    @overload
    def __call__(self, **kwargs) -> None: ...
    @overload
    def __call__[T](self, arg: T, **kwargs) -> T: ...
    @overload
    def __call__[T1, T2, *Ts](
        self, arg1: T1, arg2: T2, *args: *Ts, **kwargs
    ) -> tuple[T1, T2, *Ts]: ...

ic: IceCreamDebugger



from __future__ import annotations

import functools
import importlib.metadata
from collections.abc import Hashable
from importlib.metadata import Distribution, PackagePath
from pathlib import Path
from typing import TYPE_CHECKING

import attrs
import cachebox
import packaging.version
from packaging.version import InvalidVersion, Version

if TYPE_CHECKING:
    from _typeshed import StrPath


@attrs.define
class FilesIndex:
    _distributions: list[Distribution] = attrs.field(
        repr=False, init=False, factory=list
    )

    def add(self, distributuion: Distribution) -> None:
        self._distributions.append(distributuion)

    def has(self, file: StrPath) -> bool:
        file = Path(file).resolve()
        return self._has(str(file))

    _has_cache: cachebox.BaseCacheImpl[Hashable, bool] = attrs.field(
        repr=False, init=False, factory=lambda: cachebox.LRUCache(maxsize=1024)
    )

    @cachebox.cached(lambda self: self._has_cache)
    def _has(self, file: str) -> bool:
        if file in self._files:
            return True
        file: Path = Path(file)
        return any(file.is_relative_to(prefix) for prefix in self._pth)

    @functools.cached_property
    def _files(self) -> set[str]:
        files: set[str]
        files, _ = self._files_pth
        return files

    @functools.cached_property
    def _pth(self) -> set[str]:
        pth: set[str]
        _, pth = self._files_pth
        return pth

    @functools.cached_property
    def _files_pth(self) -> tuple[set[str], set[str]]:
        files: set[str] = set()
        pth: set[str] = set()
        for distribution in self._distributions:
            dist_files: list[PackagePath] | None = distribution.files
            if dist_files is None:
                continue
            for dist_file in dist_files:
                if dist_file.suffix == ".pth":
                    for line in dist_file.read_text().splitlines():
                        folder = Path(line)
                        if folder.is_dir():
                            pth.add(line)
                else:
                    files.add(str(dist_file.locate()))
        return files, pth


@attrs.define
class ReleaseTypeIndex:
    def is_dev(self, file: StrPath | None = None, name: str | None = None) -> bool:
        if name == "__main__":
            return True
        return file is not None and self._dev_index.has(file)

    def is_pre(
        self, file: StrPath | None = None, name: str | None = None
    ) -> bool | None:
        if name == "__main__":
            return True
        return file is not None and self._pre_index.has(file)

    @functools.cached_property
    def _dev_index(self) -> FilesIndex:
        dev_index: FilesIndex
        dev_index, _ = self._index
        return dev_index

    @functools.cached_property
    def _pre_index(self) -> FilesIndex:
        pre_index: FilesIndex
        _, pre_index = self._index
        return pre_index

    @functools.cached_property
    def _index(self) -> tuple[FilesIndex, FilesIndex]:
        dev_index: FilesIndex = FilesIndex()
        pre_index: FilesIndex = FilesIndex()
        for distribution in importlib.metadata.distributions():
            try:
                version: Version = packaging.version.parse(distribution.version)
            except InvalidVersion:
                continue
            if version.is_devrelease:
                dev_index.add(distribution)
            if version.is_prerelease:
                pre_index.add(distribution)
        return dev_index, pre_index


def is_dev_release(file: StrPath | None = None, name: str | None = None) -> bool:
    return _release_type_index.is_dev(file, name)


def is_pre_release(file: StrPath | None = None, name: str | None = None) -> bool | None:
    return _release_type_index.is_pre(file, name)


_release_type_index = ReleaseTypeIndex()



from typing import Any

import attrs


@attrs.define
class MatchError(ValueError):
    value: Any
    typ: str | type = "match"

    def __str__(self) -> str:
        cls: str = self.typ if isinstance(self.typ, str) else self.typ.__qualname__
        return f"{self.value!r} is not a valid {cls}."



from typing import Any, Unpack

import fieldz
import wadler_lindig as wl

from liblaf.grapes.sentinel import MISSING
from liblaf.grapes.wadler_lindig._typing import WadlerLindigOptions

from ._type import pdoc_type


def pdoc_fieldz(
    obj: object, **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc | None:
    try:
        fields: tuple[fieldz.Field, ...] = fieldz.fields(obj)
    except TypeError:
        return None
    cls: type = type(obj)
    pairs: list[tuple[str, Any]] = []
    for field in fields:
        if not field.repr:
            continue
        value: Any = getattr(obj, field.name, MISSING)
        if kwargs.get("hide_defaults", True) and value is field.default:
            continue
        pairs.append((field.name, value))
    return wl.bracketed(
        begin=pdoc_type(cls, dataclass=True, **kwargs) + wl.TextDoc("("),
        docs=wl.named_objs(pairs, **kwargs),
        sep=wl.comma,
        end=wl.TextDoc(")"),
        indent=kwargs.get("indent", 2),
    )



from collections.abc import Callable
from typing import Any, TypedDict

import wadler_lindig as wl

type CustomCallable = Callable[[Any], wl.AbstractDoc | None]


class WadlerLindigOptions(TypedDict, total=False):
    custom: CustomCallable
    hide_defaults: bool
    indent: int
    respect_pdoc: bool
    short_arrays: bool
    show_dataclass_module: bool
    show_type_module: bool
    width: int



import sys
import types
from typing import Any

_ARRAY_KINDS: list[tuple[str, str]] = [
    ("numpy", "ndarray"),
    ("jax", "Array"),
    ("torch", "Tensor"),
    ("cupy", "ndarray"),
]


def array_kind(obj: Any) -> str | None:
    # ref: <https://github.com/arogozhnikov/einops/blob/43a12ee010a844f3ad57bf404d27e4ee2d151131/einops/_backends.py>
    # ref: <https://github.com/patrick-kidger/wadler_lindig/blob/03ed125c04766008fb3f2eaa611fd3627b09de3d/wadler_lindig/_definitions.py#L254-L265>
    for module_name, type_name in _ARRAY_KINDS:
        module: types.ModuleType | None = sys.modules.get(module_name)
        if module is None:
            continue
        if isinstance(obj, getattr(module, type_name)):
            return module_name
    return None


def is_array(obj: Any) -> bool:
    return array_kind(obj) is not None



# ruff: noqa: A004

from ._config import BaseConfig
from ._constants import METADATA_KEY
from ._entry import Entry
from ._field import Field
from ._field_method import FieldMethod, ListFieldMethod, VarEntry
from ._field_specifiers import bool, decimal, float, int, list, path, str
from ._group import GroupEntry, group

__all__ = [
    "METADATA_KEY",
    "BaseConfig",
    "Entry",
    "Field",
    "FieldMethod",
    "GroupEntry",
    "ListFieldMethod",
    "VarEntry",
    "bool",
    "decimal",
    "float",
    "group",
    "int",
    "list",
    "path",
    "str",
]



from typing import Any, overload

import liblaf.grapes.functools as ft
from liblaf.grapes.sentinel import MISSING

from ._base import BaseTimer


@overload
def get_timer(wrapper: Any) -> BaseTimer: ...
@overload
def get_timer[T](wrapper: Any, default: T) -> BaseTimer | T: ...
def get_timer(wrapper: Any, default: Any = MISSING) -> Any:
    if default is MISSING:
        return ft.wrapt_getattr(wrapper, "timer")
    return ft.wrapt_getattr(wrapper, "timer", default)


def set_timer(wrapper: Any, timer: BaseTimer) -> None:
    ft.wrapt_setattr(wrapper, "timer", timer)



import logging
import types
from collections.abc import Generator, Iterable
from typing import cast

from rich.console import Console, ConsoleRenderable, RenderableType, RichCast
from rich.highlighter import Highlighter, ReprHighlighter
from rich.text import Text

from liblaf.grapes import pretty
from liblaf.grapes._config import config
from liblaf.grapes.rich._get_console import get_console
from liblaf.grapes.rich.traceback import RichExceptionSummary

from .columns import (
    RichHandlerColumn,
    RichHandlerColumnLevel,
    RichHandlerColumnLocation,
    RichHandlerColumnTime,
)


def _default_columns(*, time_relative: bool | None = None) -> list[RichHandlerColumn]:
    if time_relative is None:
        time_relative = config.logging.time_relative.get()
    return [
        RichHandlerColumnTime(relative=time_relative),
        RichHandlerColumnLevel(),
        RichHandlerColumnLocation(),
    ]


class RichHandler(logging.Handler):
    columns: list[RichHandlerColumn]
    console: Console
    highlighter: Highlighter

    def __init__(
        self,
        console: Console | None = None,
        *,
        columns: Iterable[RichHandlerColumn] | None = None,
        level: int = logging.NOTSET,
        time_relative: bool | None = None,
    ) -> None:
        super().__init__(level=level)
        columns = (
            _default_columns(time_relative=time_relative)
            if columns is None
            else list(columns)
        )
        if console is None:
            console = get_console(stderr=True)
        self.columns = columns
        self.console = console
        self.highlighter = ReprHighlighter()

    def emit(self, record: logging.LogRecord) -> None:
        self.console.print(
            *self._render(record),
            sep="",
            end="",
            overflow="ignore",
            no_wrap=True,
            highlight=False,
            crop=False,
            soft_wrap=False,
        )
        if (exception := self._render_exception(record)) is not None:
            self.console.print(exception)

    def _render(self, record: logging.LogRecord) -> Generator[RenderableType]:
        columns: list[Text] = [
            result
            for column in self.columns
            if (result := column.render(record)) is not None
        ]
        meta: Text = Text(" ").join(columns)
        message: Text = self._render_message(record)
        for line in message.split() or [""]:
            yield meta
            if len(line) > 0:
                yield " "
                yield line
            yield "\n"

    def _render_exception(
        self, record: logging.LogRecord
    ) -> RichExceptionSummary | None:
        if record.exc_info is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = record.exc_info
        if exc_type is None or exc_value is None:
            return None
        return RichExceptionSummary(exc_type, exc_value, traceback)

    def _render_message(self, record: logging.LogRecord) -> Text:
        if markup := getattr(record, "markup", None):
            if callable(markup):
                markup = cast("str", markup())
            return Text.from_markup(markup, style="log.message")
        if isinstance(record.msg, (ConsoleRenderable, RichCast)):
            with self.console.capture() as capture:
                self.console.print(record.msg)
            return Text.from_ansi(capture.get(), style="log.message")
        message: str = record.getMessage()
        if pretty.has_ansi(message):
            return Text.from_ansi(message, style="log.message")
        text: Text = Text(message, style="log.message")
        return self.highlighter(text)



from collections.abc import Callable
from typing import Any, TypedDict

import pydantic

type DecHook = Callable[[type, Any], Any]


class PydanticValidateOptions(TypedDict, total=False):
    strict: bool | None
    """Whether to enforce types strictly."""

    from_attributes: bool | None
    """Whether to extract data from object attributes."""

    context: Any | None
    """Additional context to pass to the validator."""

    by_alias: bool | None
    """Whether to use the field's alias when validating against the provided input data."""

    by_name: bool | None
    """Whether to use the field's name when validating against the provided input data."""


def dec_hook(
    typ: type, obj: Any, /, *, pydantic_options: PydanticValidateOptions | None = None
) -> Any:
    if issubclass(typ, pydantic.BaseModel):
        pydantic_options = pydantic_options or {}
        return typ.model_validate(obj, **pydantic_options)
    return typ(obj)



""".

Examples:
    >>> import attrs
    >>> @define
    ... class SomeClass:
    ...     a_number: int = 42
    ...     list_of_numbers: list[int] = attrs.Factory(list)
    >>> obj = SomeClass()
    >>> obj
    SomeClass(list_of_numbers=[])
"""

import functools
from typing import Any

import attrs

from liblaf.grapes.functools import wraps
from liblaf.grapes.rich.repr import rich_repr_fieldz
from liblaf.grapes.wadler_lindig import pdoc_fieldz, pdoc_rich_repr, pformat


@wraps(attrs.define)
def define(maybe_cls: type | None = None, **kwargs) -> Any:
    if maybe_cls is None:
        return functools.partial(define, **kwargs)
    cls: type = _preprocess(maybe_cls, kwargs)
    return attrs.define(cls, **kwargs)


@wraps(attrs.frozen)
def frozen(maybe_cls: type | None = None, **kwargs) -> Any:
    if maybe_cls is None:
        return functools.partial(frozen, **kwargs)
    cls: type = _preprocess(maybe_cls, kwargs)
    return attrs.frozen(cls, **kwargs)


def _preprocess[T: type](cls: T, kwargs: dict[str, Any]) -> T:
    auto_detect: bool = kwargs.get("auto_detect", True)
    repr_: bool | None = kwargs.get("repr")
    if auto_detect and repr_ is None:
        repr_ = cls.__repr__ is object.__repr__
    if repr_:
        cls.__repr__ = pformat  # pyright: ignore[reportAttributeAccessIssue]
    kwargs["repr"] = False
    if not hasattr(cls, "__pdoc__"):
        if hasattr(cls, "__rich_repr__"):
            cls.__pdoc__ = pdoc_rich_repr
        else:
            cls.__pdoc__ = pdoc_fieldz
    if not hasattr(cls, "__rich_repr__"):
        cls.__rich_repr__ = rich_repr_fieldz
    return cls



def has_ansi(s: str, /) -> bool:
    r""".

    Examples:
        >>> has_ansi("\x1b[31m red text \x1b[0m")
        True
        >>> has_ansi("plain text")
        False
    """
    return "\x1b" in s



from __future__ import annotations

import datetime
import enum
import functools
import types
from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING, Any, Unpack, overload

import pydantic
import wadler_lindig as wl

from liblaf.grapes.wadler_lindig._typing import CustomCallable, WadlerLindigOptions

from ._array import pdoc_array
from ._datetime import pdoc_datetime
from ._enum import pdoc_enum
from ._fieldz import pdoc_fieldz
from ._mapping import pdoc_mapping
from ._pydantic import pdoc_pydantic_root_model
from ._rich_repr import pdoc_rich_repr

if TYPE_CHECKING:
    from functools import _RegType, _SingleDispatchCallable


class PdocCustomDispatcher:
    _dispatcher: _SingleDispatchCallable[wl.AbstractDoc | None]
    _registry: list[CustomCallable]

    def __init__(self) -> None:
        @functools.singledispatch
        def dispatcher(
            obj: Any, **kwargs: Unpack[WadlerLindigOptions]
        ) -> wl.AbstractDoc | None:
            return self._fallback(obj, **kwargs)

        self._dispatcher = dispatcher
        self._registry = []
        self.register(datetime.date, pdoc_datetime)
        self.register(enum.Enum, pdoc_enum)
        self.register(Mapping, pdoc_mapping)
        self.register(pydantic.RootModel, pdoc_pydantic_root_model)
        self.register(pdoc_array)
        self.register(pdoc_fieldz)
        self.register(pdoc_rich_repr)

    def __call__(
        self, obj: Any, **kwargs: Unpack[WadlerLindigOptions]
    ) -> wl.AbstractDoc | None:
        respect_pdoc: bool = kwargs.get("respect_pdoc", True)
        if respect_pdoc and hasattr(obj, "__pdoc__"):
            return None
        return self._dispatcher(obj, **kwargs)

    @overload
    def register[C: Callable](self, cls: _RegType, /) -> Callable[[C], C]: ...
    @overload
    def register[C: Callable](self, cls: _RegType, /, func: C) -> C: ...
    @overload
    def register[C: Callable](self, func: C, /) -> C: ...
    def register(
        self, cls_or_func: _RegType | Callable, /, func: Callable | None = None
    ) -> Callable:
        if isinstance(cls_or_func, (type, types.UnionType)):
            return self._dispatcher.register(cls_or_func, func)
        self._registry.append(cls_or_func)
        return cls_or_func

    def _fallback(self, obj: Any, **kwargs: Any) -> wl.AbstractDoc | None:
        for func in self._registry:
            doc: wl.AbstractDoc | None = func(obj, **kwargs)
            if doc is not None:
                return doc
        return None


pdoc_custom: PdocCustomDispatcher = PdocCustomDispatcher()



import types
from collections.abc import Sequence
from typing import Any

type ClassInfo = type | types.UnionType | tuple[ClassInfo, ...]


def as_sequence(obj: Any, base_type: ClassInfo | None = (str, bytes)) -> Sequence:
    """.

    Examples:
        If `obj` is iterable, return an iterator over its items:

        >>> obj = (1, 2, 3)
        >>> as_sequence(obj)
        (1, 2, 3)

        If `obj` is not iterable, return a one-item iterable containing `obj`:

        >>> obj = 1
        >>> as_sequence(obj)
        (1,)

        If `obj` is `None`, return an empty iterable:

        >>> obj = None
        >>> as_sequence(None)
        ()

        By default, binary and text strings are not considered iterable:

        >>> obj = "foo"
        >>> as_sequence(obj)
        ('foo',)

        If `base_type` is set, objects for which `isinstance(obj, base_type)` returns ``True`` won't be considered iterable.

        >>> obj = {"a": 1}
        >>> as_sequence(obj)
        ({'a': 1},)
        >>> as_sequence(obj, base_type=dict)  # Treat dicts as a unit
        ({'a': 1},)

        Set `base_type` to `None` to avoid any special handling and treat objects Python considers iterable as iterable:

        >>> obj = "foo"
        >>> as_sequence(obj, base_type=None)
        'foo'

    References:
        1. [`more_itertools.always_iterable`](https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.always_iterable)
    """
    if obj is None:
        return ()
    if base_type is not None and isinstance(obj, base_type):
        return (obj,)
    if isinstance(obj, Sequence):
        return obj
    return (obj,)



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath


def abbr_path(path: StrPath, truncation_symbol: str = "󰇘/") -> str:
    path = Path(path)
    for prefix in sys.path:
        if path.is_relative_to(prefix):
            return f"{truncation_symbol}{path.relative_to(prefix)}"
    return str(path)



from . import (
    attrs,
    bench,
    compat,
    conf,
    deps,
    errors,
    functools,
    icecream,
    itertools,
    logging,
    magic,
    pretty,
    rich,
    sentinel,
    serde,
    timing,
    typing,
    wadler_lindig,
    warnings,
)
from ._config import config
from ._version import __commit_id__, __version__, __version_tuple__
from .bench import Bencher, BenchResults
from .compat import contains, getitem
from .deps import optional_deps
from .errors import (
    DispatchLookupError,
    MatchError,
    TodoError,
    UnreachableError,
    todo,
    unreachable,
)
from .functools import memorize, wraps, wrapt_getattr, wrapt_setattr
from .itertools import (
    areidentical,
    as_iterable,
    as_sequence,
    compact,
    first_not_none,
    keyjoin,
    len_or_none,
    omit,
    pick,
)
from .logging import LazyRepr, LazyStr
from .magic import entrypoint, in_ci
from .pretty import (
    has_ansi,
    pretty_call,
    pretty_duration,
    pretty_durations,
    pretty_func,
    pretty_quantities,
    pretty_quantity,
    pretty_quantity_components,
    pretty_throughput,
)
from .rich import get_console
from .rich.progress import Progress, track
from .rich.repr import auto_rich_repr
from .sentinel import MISSING, MissingType
from .serde import dec_hook, enc_hook, json, load, save, toml, yaml
from .timing import BaseTimer, Timer, get_timer, timer
from .typing import array_kind, is_array
from .wadler_lindig import pdoc, pformat, pprint
from .warnings import warn

__all__ = [
    "MISSING",
    "BaseTimer",
    "BenchResults",
    "Bencher",
    "DispatchLookupError",
    "LazyRepr",
    "LazyStr",
    "MatchError",
    "MissingType",
    "Progress",
    "Timer",
    "TodoError",
    "UnreachableError",
    "__commit_id__",
    "__version__",
    "__version_tuple__",
    "areidentical",
    "array_kind",
    "as_iterable",
    "as_sequence",
    "attrs",
    "auto_rich_repr",
    "bench",
    "compact",
    "compat",
    "conf",
    "config",
    "contains",
    "dec_hook",
    "deps",
    "enc_hook",
    "entrypoint",
    "errors",
    "first_not_none",
    "functools",
    "get_console",
    "get_timer",
    "getitem",
    "has_ansi",
    "icecream",
    "in_ci",
    "is_array",
    "itertools",
    "json",
    "keyjoin",
    "len_or_none",
    "load",
    "logging",
    "magic",
    "memorize",
    "omit",
    "optional_deps",
    "pdoc",
    "pformat",
    "pick",
    "pprint",
    "pretty",
    "pretty_call",
    "pretty_duration",
    "pretty_durations",
    "pretty_func",
    "pretty_quantities",
    "pretty_quantity",
    "pretty_quantity_components",
    "pretty_throughput",
    "rich",
    "save",
    "sentinel",
    "serde",
    "timer",
    "timing",
    "todo",
    "toml",
    "track",
    "typing",
    "unreachable",
    "wadler_lindig",
    "warn",
    "warnings",
    "wraps",
    "wrapt_getattr",
    "wrapt_setattr",
    "yaml",
]



from ._dispatch import DispatchLookupError
from ._match import MatchError
from ._misc import TodoError, UnreachableError, todo, unreachable

__all__ = [
    "DispatchLookupError",
    "MatchError",
    "TodoError",
    "UnreachableError",
    "todo",
    "unreachable",
]



from typing import Any

import fieldz
from rich.repr import RichReprResult

from liblaf.grapes.sentinel import MISSING


def rich_repr_fieldz(obj: object) -> RichReprResult:
    for field in fieldz.fields(obj):
        if not field.repr:
            continue
        value: Any = getattr(obj, field.name, MISSING)
        # rich.repr uses `if default == value:` but does not protect against
        # exceptions. Some types (e.g. NumPy arrays) raise on equality/truth
        # checks (ambiguous truth value). Do the comparison here inside a
        # try/except so we can catch and handle those errors safely.
        try:
            if value == field.default:
                yield field.name, value, field.default
            else:
                yield field.name, value
        except Exception:  # noqa: BLE001
            yield field.name, value



import functools
from collections.abc import Iterable
from typing import Any, overload

import limits

from ._statistics import StatisticName
from ._timings import Callback, Timings


@overload
def log_record(
    timer: Timings,
    /,
    *,
    index: int = ...,
    level: int | str = ...,
    limits: str | limits.RateLimitItem | None = ...,
) -> Any: ...
@overload
def log_record(
    *,
    index: int = ...,
    level: int | str = ...,
    limits: str | limits.RateLimitItem | None = ...,
) -> Callback: ...
def log_record(timer: Timings | None = None, /, **kwargs) -> Any:
    _logging_hide = True
    if timer is None:
        return functools.partial(log_record, **kwargs)
    return timer.log_record(**kwargs)


@overload
def log_summary(
    timer: Timings,
    /,
    *,
    level: int | str = ...,
    stats: Iterable[StatisticName] = ...,
    limits: str | limits.RateLimitItem | None = ...,
) -> None: ...
@overload
def log_summary(
    *,
    level: int | str = ...,
    stats: Iterable[StatisticName] = ...,
    limits: str | limits.RateLimitItem | None = ...,
) -> Callback: ...
def log_summary(timer: Timings | None = None, /, **kwargs) -> Any:
    _logging_hide = True
    if timer is None:
        return functools.partial(log_summary, **kwargs)
    return timer.log_summary(**kwargs)



import warnings

from liblaf.grapes import magic
from liblaf.grapes.functools import wraps


@wraps(warnings.warn)
def warn(*args, **kwargs) -> None:
    _warnings_hide = True
    stacklevel: int = kwargs.get("stacklevel", 1)
    _, stacklevel = magic.get_frame_with_stacklevel(
        stacklevel, magic.hidden_from_warnings
    )
    kwargs["stacklevel"] = stacklevel
    return warnings.warn(*args, **kwargs)



from liblaf.grapes import conf
from liblaf.grapes.conf import BaseConfig, Field


class ConfigWarnings(BaseConfig):
    hide_stable_release: Field[bool] = conf.bool(
        default=True, env="WARNINGS_HIDE_STABLE_RELEASE"
    )



import functools
import itertools
import linecache
import types
from collections.abc import Generator
from typing import Any

import attrs
import tlz
from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.scope import render_scope
from rich.syntax import Syntax
from rich.text import Text

from liblaf.grapes import magic
from liblaf.grapes._config import config

from ._options import RichTracebackOptions


@attrs.define
class RichFrameSummary:
    frame: types.FrameType
    lineno: int
    lasti: int

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield from self.render()

    @functools.cached_property
    def filename(self) -> str:
        return self.frame.f_code.co_filename

    @functools.cached_property
    def hidden(self) -> bool:
        return magic.hidden_from_traceback(self.frame)

    @functools.cached_property
    def locals(self) -> dict[str, Any]:
        return self.frame.f_locals

    @functools.cached_property
    def name(self) -> str:
        return self.frame.f_code.co_name

    @functools.cached_property
    def position(self) -> tuple[int | None, int | None, int | None, int | None]:
        return _get_code_position(self.frame.f_code, self.lasti)

    @functools.cached_property
    def qualname(self) -> str:
        return self.frame.f_code.co_qualname

    @property
    def start_line(self) -> int | None:
        return self.position[0]

    @property
    def end_line(self) -> int | None:
        return self.position[1]

    @property
    def start_column(self) -> int | None:
        return self.position[2]

    @property
    def end_column(self) -> int | None:
        return self.position[3]

    def render(
        self, options: RichTracebackOptions | None = None
    ) -> Generator[RenderableType]:
        if options is None:
            options = RichTracebackOptions()
        yield from self._render_location(options)
        if not self.hidden:
            yield from self._render_syntax(options)
            if options.show_locals:
                yield from self._render_locals(options)

    def _render_location(
        self, _options: RichTracebackOptions
    ) -> Generator[RenderableType]:
        filename: str = magic.abbr_path(self.filename)
        qualname: str = self.qualname
        if qualname != "<module>":
            qualname += "()"
        if self.hidden:
            # ? I don't know why, but adding "white" looks better on Ghostty.
            text: Text = Text.assemble(
                (filename, "repr.filename"),
                (":", "white"),
                (str(self.lineno), "repr.number"),
                (" in ", "white"),
                (qualname, "repr.call"),
                (" --- hidden", "white"),
                style="dim",
            )
            yield text
        else:
            yield Text.assemble(
                (filename, "repr.filename"),
                ":",
                (str(self.lineno), "repr.number"),
                " in ",
                (qualname, "repr.call"),
            )

    def _render_syntax(
        self, options: RichTracebackOptions
    ) -> Generator[RenderableType]:
        lines: list[str] = linecache.getlines(self.filename, self.frame.f_globals)
        code: str = "".join(lines)
        lexer: str = Syntax.guess_lexer(self.filename, code)
        syntax = Syntax(
            code,
            lexer,
            theme=options.theme,
            line_numbers=True,
            line_range=(self.start_line, self.end_line),
            indent_guides=True,
        )
        if all(v is not None for v in self.position):
            # ref: <https://github.com/Textualize/rich/blob/4d6d631a3d2deddf8405522d4b8c976a6d35726c/rich/traceback.py#L841C25-L859>
            # Stylize a line at a time
            # So that indentation isn't underlined (which looks bad)
            for lineno, start_col, end_col in _iter_syntax_lines(*self.position):  # pyright: ignore[reportArgumentType]
                if start_col == 0:
                    line: str = lines[lineno - 1]
                    stripped: str = line.lstrip()
                    start_col: int = len(line) - len(stripped)  # noqa: PLW2901
                if end_col == -1:
                    end_col: int = len(lines[lineno - 1])  # noqa: PLW2901
                syntax.stylize_range(
                    "traceback.error_range", (lineno, start_col), (lineno, end_col)
                )
        yield syntax

    def _render_locals(
        self, options: RichTracebackOptions
    ) -> Generator[RenderableType]:
        locals_: dict[str, Any] = self.locals
        if options.locals_hide_dunder:
            locals_ = tlz.keyfilter(_not_dunder, locals_)
        if options.locals_hide_sunder:
            locals_ = tlz.keyfilter(_not_sunder, locals_)
        if not locals_:
            return
        with config.pretty.indent.overrides(4):
            yield render_scope(
                locals_,
                title="locals",
                indent_guides=options.indent_guide,
                max_length=options.locals_max_length,
                max_string=options.locals_max_string,
            )


def _not_dunder(key: str) -> bool:
    return not key.startswith("__")


def _not_sunder(key: str) -> bool:
    return not key.startswith("_") or key.startswith("__")


def _get_code_position(
    code: types.CodeType, instruction_index: int
) -> tuple[int | None, int | None, int | None, int | None]:
    # ref: <https://github.com/python/cpython/blob/4885ecfbda4cc792691e5d488ef6cb09727eb417/Lib/traceback.py#L431-L435>
    if instruction_index < 0:
        return (None, None, None, None)
    return next(itertools.islice(code.co_positions(), instruction_index // 2, None))


def _iter_syntax_lines(
    start_line: int, end_line: int, start_column: int, end_column: int
) -> Generator[tuple[int, int, int]]:
    # ref: <https://github.com/Textualize/rich/blob/4d6d631a3d2deddf8405522d4b8c976a6d35726c/rich/traceback.py#L55-L80>
    if start_line == end_line:
        yield start_line, start_column, end_column
    else:
        yield start_line, start_column, -1
        for line in range(start_line + 1, end_line):
            yield line, 0, -1
        yield end_line, 0, end_column



import functools
import logging
from collections.abc import Iterable, Sequence
from typing import Any, overload

import attrs
import limits


@attrs.define
class LimitsFilter:
    limiter: limits.strategies.RateLimiter = attrs.field(
        factory=lambda: limits.strategies.FixedWindowRateLimiter(
            limits.storage.MemoryStorage()
        )
    )

    def filter(self, record: logging.LogRecord) -> bool:
        args: Any = getattr(record, "limits", None)
        if args is None:
            return True
        hit_args: LimitsHitArgs = _parse_args_cache(args)
        if hit_args.item is None:
            return True
        identifiers: Sequence[str] = hit_args.make_identifiers(record)
        return self.limiter.hit(hit_args.item, *identifiers, cost=hit_args.cost)


@overload
def _parse_item(item: None) -> None: ...
@overload
def _parse_item(item: str | limits.RateLimitItem) -> limits.RateLimitItem: ...
def _parse_item(item: str | limits.RateLimitItem | None) -> limits.RateLimitItem | None:
    match item:
        case None:
            return None
        case str():
            return limits.parse(item)
        case limits.RateLimitItem():
            return item
        case _:
            raise TypeError(item)


@attrs.frozen
class LimitsHitArgs:
    item: limits.RateLimitItem | None = attrs.field(converter=_parse_item)
    namespace: Iterable[str] | None = attrs.field(default=None)
    identifiers: Iterable[str] = attrs.field(default=())
    cost: int = 1

    def make_identifiers(self, record: logging.LogRecord) -> Sequence[str]:
        namespace: Iterable[str] = (
            self._default_namespace(record)
            if self.namespace is None
            else self.namespace
        )
        return (*namespace, *self.identifiers)

    @staticmethod
    def _default_namespace(record: logging.LogRecord) -> Sequence[str]:
        return record.pathname, str(record.lineno), record.levelname


@functools.singledispatch
def _parse_args(args: Any) -> LimitsHitArgs:
    raise ValueError(args)


_parse_args_cache = functools.lru_cache(_parse_args)


@_parse_args.register(str | limits.RateLimitItem)
def _(args: str | limits.RateLimitItem) -> LimitsHitArgs:
    return LimitsHitArgs(item=args)


@_parse_args.register(LimitsHitArgs)
def _(args: LimitsHitArgs) -> LimitsHitArgs:
    return args



from pathlib import Path

import platformdirs

from liblaf.grapes import conf
from liblaf.grapes.conf import BaseConfig, Field


def _default_location() -> Path:
    return platformdirs.user_cache_path(appname="joblib")


class ConfigJoblibMemory(BaseConfig):
    bytes_limit: Field[str] = conf.str(default="4G")
    location: Field[Path] = conf.path(factory=_default_location)


class ConfigJoblib(BaseConfig):
    memory: ConfigJoblibMemory = conf.group()



from logging import LogRecord
from typing import override

from rich.text import Text

from ._abc import RichHandlerColumn


class RichHandlerColumnLocation(RichHandlerColumn):
    @override
    def render(self, record: LogRecord) -> Text | None:
        if record.name == "py.warnings":
            return Text(record.name, "log.path")
        plain: str = f"{record.name}:{record.funcName}:{record.lineno}"
        return Text(plain, "log.path")



from ._ansi import has_ansi
from ._call import pretty_call
from ._duration import pretty_duration, pretty_durations
from ._func import pretty_func
from ._quantiphy import (
    PrettyQuantitiesComponents,
    PrettyQuantityComponents,
    pretty_quantities,
    pretty_quantity,
    pretty_quantity_components,
)
from ._throughput import pretty_throughput
from ._utils import get_name

__all__ = [
    "PrettyQuantitiesComponents",
    "PrettyQuantityComponents",
    "get_name",
    "has_ansi",
    "pretty_call",
    "pretty_duration",
    "pretty_durations",
    "pretty_func",
    "pretty_quantities",
    "pretty_quantity",
    "pretty_quantity_components",
    "pretty_throughput",
]



import functools
import os
import sys
from typing import IO

import rich
from rich.console import Console
from rich.style import Style
from rich.theme import Theme

from liblaf.grapes import magic
from liblaf.grapes.functools import wraps


def default_theme() -> Theme:
    """.

    References:
        1. <https://github.com/Delgan/loguru/blob/master/loguru/_defaults.py>
    """
    return Theme(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(color="cyan", bold=True),
            "logging.level.debug": Style(color="blue", bold=True),
            "logging.level.icecream": Style(color="magenta", bold=True),
            "logging.level.info": Style(bold=True),
            "logging.level.success": Style(color="green", bold=True),
            "logging.level.warning": Style(color="yellow", bold=True),
            "logging.level.error": Style(color="red", bold=True),
            "logging.level.critical": Style(color="red", bold=True, reverse=True),
        },
        inherit=True,
    )


@wraps(Console)
@functools.cache
def get_console(**kwargs) -> Console:
    if kwargs.get("theme") is None:
        kwargs["theme"] = default_theme()
    file: IO[str] | None = kwargs.get("file")
    stderr: bool = file is None and kwargs.get("stderr", False)
    stdout: bool = file is None and not stderr
    if "force_terminal" not in kwargs and (stdout or stderr) and magic.in_ci():
        kwargs["force_terminal"] = True
    if "width" not in kwargs and (
        (stdout and not sys.stdout.isatty())
        or (stderr and not sys.stderr.isatty())
        or (file is not None and not os.isatty(file.fileno()))
    ):
        kwargs["width"] = 128
    if stdout:
        rich.reconfigure(**kwargs)
        return rich.get_console()
    return Console(**kwargs)
